from __future__ import annotations
from math import exp, log10, log, pow
from typing import Callable


def db_to_volume_value(db: float):
    '''
    Maps a dB value to a volume value (0 - 1).
    '''
    return exp((db - 6) * (1 / 20)) if db > -100 else 0


def gain_to_db(gain: float):
    '''
    @param gain A value between 0 - 2, corresponding to -inf to +6dB, gain == 1 equals dB == 0.0
    '''
    if (gain <= 0):
        return -100

    return 20 * log10(gain)


def volume_value_to_db(volume: float):
    '''
    @param volume A value between 0 - 1, corresponding to -inf to +6 dB.
    '''
    return 20 * log(volume) + 6 if volume > 0 else -100.0


def volume_value_to_gain(volume: float):
    '''
    @param volume A value between 0 - 1, corresponding to -inf to +6 dB.
    '''
    return pow(10, (20 * log(volume) + 6) * (1 / 20)) if volume > 0 else 0


def gain_to_volume_value(gain: float):
    '''
    @param gain A value between 0 - 2, corresponding to -inf to +6dB, gain == 1 equals dB == 0.0
    '''
    return exp((20 * log10(gain) - 6) * (1 / 20)) if gain > 0 else 0


def greater_equal(sorted_list: list, val, key: Callable | None = None, low: int | None = None, high: int | None = None):
    '''
    Returns the index of the first item in the array >= val. This is a successor query which also returns the item if present.
    '''
    low = low if low is not None else 0
    high = high if high is not None else len(sorted_list) - 1
    i = high + 1
    while low <= high:
        m = (low + high) >> 1
        x = sorted_list[m]
        p = (key(x) - key(val)) if key is not None else (x - val)
        if p >= 0:
            i = m
            high = m - 1
        else:
            low = m + 1
    return i


def greater_than(sorted_list: list, val, key: Callable | None = None, low: int | None = None, high: int | None = None):
    '''
    Returns the index of the first item in the array > val. This is the same as a successor query.
    '''
    low = low if low is not None else 0
    high = high if high is not None else len(sorted_list) - 1
    i = high + 1
    while low <= high:
        m = (low + high) >> 1
        x = sorted_list[m]
        p = (key(x) - key(val)) if key is not None else (x - val)
        if p > 0:
            i = m
            high = m - 1
        else:
            low = m + 1
    return i


def lower_than(sorted_list: list, val, key: Callable | None = None, low: int | None = None, high: int | None = None):
    '''
    Returns the index of the last item in the array < val. This is the same as a predecessor query.
    '''
    low = low if low is not None else 0
    high = high if high is not None else len(sorted_list) - 1
    i = low - 1
    while low <= high:
        m = (low + high) >> 1
        x = sorted_list[m]
        p = (key(x) - key(val)) if key is not None else (x - val)
        if p < 0:
            i = m
            low = m + 1
        else:
            high = m - 1
    return i


def lower_equal(sorted_list: list, val, key: Callable | None = None, low: int | None = None, high: int | None = None):
    '''
    Returns the index of the last item in the array <= val. This is a predecessor query which also returns the item if present.
    '''
    low = low if low is not None else 0
    high = high if high is not None else len(sorted_list) - 1
    i = low - 1
    while low <= high:
        m = (low + high) >> 1
        x = sorted_list[m]
        p = (key(x) - key(val)) if key is not None else (x - val)
        if (p <= 0):
            i = m
            low = m + 1
        else:
            high = m - 1
    return i


def eq(sorted_list: list, val, key: Callable | None = None, low: int | None = None, high: int | None = None):
    '''
    Returns an index of some item in the array == y or -1 if the item is not present.
    '''
    low = low if low is not None else 0
    high = high if high is not None else len(sorted_list) - 1
    while low <= high:
        m = (low + high) >> 1
        x = sorted_list[m]
        p = (key(x) - key(val)) if key is not None else (x - val)
        if p == 0:
            return m
        if p <= 0:
            low = m + 1
        else:
            high = m - 1
    return -1


def get_audio_plugin_tuneflow_id(
    manufacturer_name: str,
    plugin_format_name: str,
    plugin_name: str,
    plugin_version: str,
):
    '''
    Gets an id that Tuneflow can uniquely identify a plugin.
    '''
    return f'{manufacturer_name} // {plugin_format_name} // {plugin_name} // {plugin_version}'


def get_audio_plugin_versionless_tuneflow_id(
    manufacturer_name: str,
    plugin_format_name: str,
    plugin_name: str,
):
    '''
    Gets an id that Tuneflow can uniquely identify a plugin regardless of version.
    '''
    return f'{manufacturer_name} // {plugin_format_name} // {plugin_name}'


def decode_audio_plugin_tuneflow_id(tf_id: str):
    parts = tf_id.split(' // ')
    if len(parts) < 4:
        raise Exception('Invalid audio plugin tuneflow id.')

    return {
        "name": parts[2],
        "manufacturer_name": parts[0],
        "plugin_format_name": parts[1],
        "plugin_version": parts[3],
    }


def to_versionless_tf_id(tf_id: str):
    '''
    Converts a full tf_id to a versionless tf_id.
    @param tf_id A full tf_id
    @returns A versionless tf_id
    '''
    parts = decode_audio_plugin_tuneflow_id(tf_id)
    return get_audio_plugin_versionless_tuneflow_id(
        parts["manufacturer_name"],
        parts["plugin_format_name"],
        parts["name"],
    )


def are_tuneflow_ids_equal(tf_id1: str, tf_id2: str):
    return tf_id1 == tf_id2


def are_tuneflow_ids_equal_ignore_version(tf_id1: str, tf_id2: str):
    parsed_id1 = decode_audio_plugin_tuneflow_id(tf_id1)
    parsed_id2 = decode_audio_plugin_tuneflow_id(tf_id2)
    if len(parsed_id1) != len(parsed_id2):
        return False

    for key in parsed_id1:
        if (key == 'plugin_version'):
            continue
        if key not in parsed_id2:
            return False
        if parsed_id1[key] != parsed_id2[key]:
            return False
    return True

def pitch_to_hz(pitch: int):
  '''
  @param pitch A number from 0 to 127.
  '''
  return 440 * pow(2, (pitch - 69) / 12)


from __future__ import annotations
from tuneflow_py.models.protos import song_pb2

StructureType = song_pb2.StructureMarker.StructureType


class StructureMarker:
    def __init__(self, song, tick:int=None, type: int | None = None,
                 custom_name: str | None = None, proto: song_pb2.StructureMarker | None = None) -> None:
        if song is None:
            raise Exception('song must be provided when creating a track')
        self.song = song
        if proto is not None:
            self._proto = proto
            return
        self._proto = song_pb2.StructureMarker()
        self._proto.type = type if type is not None else StructureType.INTRO
        self._proto.tick = tick if tick is not None else 0
        if type == StructureType.CUSTOM:
            self._proto.custom_name = custom_name if custom_name is not None else ''

    def get_tick(self) -> int:
        return self._proto.tick

    def set_tick(self, tick: int):
        self._proto.tick = tick

    def get_type(self) -> StructureType:
        return self._proto.type

    def set_type(self, type: StructureType):
        self._proto.type = type
    
    def set_custom_name(self, name: str):
        self._proto.custom_name = name
    
    def get_custom_name(self) -> str:
        return self._proto.custom_name


from __future__ import annotations
from tuneflow_py.models.protos import song_pb2


class TimeSignatureEvent:
    def __init__(self, ticks: int | None = None, numerator: int | None = None, denominator: int | None = None,
                 proto: song_pb2.TimeSignatureEvent | None = None):
        if proto is not None:
            self._proto = proto
        else:
            self._proto = song_pb2.TimeSignatureEvent()
            self._proto.ticks = ticks
            self._proto.numerator = numerator
            self._proto.denominator = denominator

    def get_ticks(self) -> int:
        return self._proto.ticks

    def set_ticks(self, ticks: int):
        self._proto.ticks = ticks

    def get_numerator(self) -> int:
        return self._proto.numerator

    def set_numerator(self, numerator: int):
        self._proto.numerator = numerator

    def get_denominator(self) -> int:
        return self._proto.denominator

    def set_denominator(self, denominator: int):
        self._proto.denominator = denominator

    def __repr__(self) -> str:
        return str(self._proto)



from typing_extensions import Required, NotRequired, TypedDict


class AudioData(TypedDict):
    '''
    A temporary holder of audio data that will be converted to local files and cleared once received.
    '''

    format: Required[str]
    data: Required[bytes]


class AudioClipData(TypedDict):
    ''' Audio-related clip data. '''

    audio_file_path: NotRequired[str]
    '''
    The absolute file path of the audio.
    
    Can be omitted if `audioData` is set.
    '''

    audio_data: NotRequired[AudioData]
    '''
    A temporary holder of the audio file data. When set, the daw will save a copy to local file system and replace
    `audioFilePath` with the saved file path, and this field will be cleared afterwards.
    
    Can be omitted if `audioFilePath` is set.

    IMPORTANT: Only supported on desktop, requires desktop app version >= 1.8.9.
    '''

    start_tick: Required[int]
    '''
    The start tick of the audio.
    
    This is an absolute position that locates the start of the audio within the song.
    
    When the clip adjusts left or right boundaries, the start position of the audio
    will not change.
    
    When the clip moves, the audio will move along with the clip.
    '''

    duration: Required[float]
    '''
    The duration of the audio content. You can get this value in the plugin
    by using `readApis.readAudioBuffer` to read the audio file content as `AudioBuffer`, and then
    get the duration from `audioBuffer.duration`.
    
    Duration needs to be updated whenever audio file (path or content) changes.
    '''

    speed_ratio: NotRequired[float]
    '''
    How fast this audio plays, time-stretch is applied when this ratio is set and not 1, default to 1.
    '''

    pitch_offset: NotRequired[float]
    '''
    Pitch offset in semitones, ranging from -48 to 48, default to 0.
    '''


from tuneflow_py import Song, TrackType, ClipType
import unittest
import pytest

TEST_AUDIO_CLIP_DATA = {
    "audio_file_path": 'file1',
    "start_tick": 480,
    "duration": 1,
}

TEST_AUDIO_CLIP_DATA_WITH_AUDIO_DATA = {
    "audio_data": {
        "data": bytes(),
        "format": "wav"
    },
    "start_tick": 480,
    "duration": 1,
}


def create_song():
    song = Song()
    song.create_tempo_change(ticks=1440, bpm=60)
    audio_track = song.create_track(type=TrackType.AUDIO_TRACK)
    midi_track = song.create_track(type=TrackType.MIDI_TRACK)
    return song, audio_track, midi_track


class BaseTestCase(unittest.TestCase):
    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        [song, audio_track, midi_track] = create_song()
        self.song = song
        self.audio_track = audio_track
        self.midi_track = midi_track


class TestCreateAudioClip(BaseTestCase):
    def test_create_audio_clip_by_default(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=0,
            audio_clip_data=TEST_AUDIO_CLIP_DATA,
        )
        self.assertEqual(clip1.get_type(), ClipType.AUDIO_CLIP)
        clip2 = self.midi_track.create_midi_clip(
            clip_start_tick=0,
        )
        self.assertEqual(clip2.get_type(), ClipType.MIDI_CLIP)
        clip3 = self.audio_track.create_midi_clip(
            clip_start_tick=0,
        )
        self.assertEqual(clip3.get_type(), ClipType.MIDI_CLIP)
        clip4 = self.midi_track.create_audio_clip(
            clip_start_tick=0,
            audio_clip_data=TEST_AUDIO_CLIP_DATA,
        )
        self.assertEqual(clip4.get_type(), ClipType.AUDIO_CLIP)


class TestAudioClipCannotCreateNotes(BaseTestCase):
    def test_cannot_create_note_in_audio_clip(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=0,
            audio_clip_data=TEST_AUDIO_CLIP_DATA,
        )
        self.assertEqual(clip1.get_raw_note_count(), 0)
        new_note = clip1.create_note(
            pitch=64,
            velocity=64,
            start_tick=0,
            end_tick=100,
        )
        self.assertIsNone(new_note)
        self.assertEqual(clip1.get_raw_note_count(), 0)


class TestBasicGetAndSetOperations(BaseTestCase):
    def test_get_audio_related_fields(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=480,
            audio_clip_data=TEST_AUDIO_CLIP_DATA,
        )
        self.assertEqual(clip1.get_clip_end_tick(), 1440)
        self.assertEqual(clip1.get_audio_duration(), 1)
        self.assertEqual(clip1.get_audio_end_tick(), 1440)
        self.assertEqual(clip1.get_audio_clip_data().audio_file_path, TEST_AUDIO_CLIP_DATA["audio_file_path"])
        self.assertEqual(clip1.get_audio_clip_data().start_tick, TEST_AUDIO_CLIP_DATA["start_tick"])
        self.assertAlmostEqual(clip1.get_audio_clip_data().duration, TEST_AUDIO_CLIP_DATA["duration"])

    def test_get_audio_related_fields_with_audio_data(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=480,
            audio_clip_data=TEST_AUDIO_CLIP_DATA_WITH_AUDIO_DATA,
        )
        self.assertEqual(clip1.get_clip_end_tick(), 1440)
        self.assertEqual(clip1.get_audio_duration(), 1)
        self.assertEqual(clip1.get_audio_end_tick(), 1440)
        self.assertEqual(clip1.get_audio_clip_data().audio_file_path, "")
        self.assertEqual(clip1.get_audio_clip_data().audio_data.format,
                         TEST_AUDIO_CLIP_DATA_WITH_AUDIO_DATA["audio_data"]["format"])
        self.assertEqual(clip1.get_audio_clip_data().audio_data.data,
                         TEST_AUDIO_CLIP_DATA_WITH_AUDIO_DATA["audio_data"]["data"])
        self.assertEqual(clip1.get_audio_clip_data().start_tick, TEST_AUDIO_CLIP_DATA_WITH_AUDIO_DATA["start_tick"])
        self.assertAlmostEqual(clip1.get_audio_clip_data().duration, TEST_AUDIO_CLIP_DATA_WITH_AUDIO_DATA["duration"])

    def test_set_and_get_custom_clip_start_and_end_tick_out_of_audio_range(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=120,
            clip_end_tick=9000,
            audio_clip_data=TEST_AUDIO_CLIP_DATA,
        )
        self.assertEqual(clip1.get_clip_start_tick(), 480)
        self.assertEqual(clip1.get_clip_end_tick(), 1440)
        self.assertEqual(clip1.get_audio_duration(), 1)
        self.assertEqual(clip1.get_audio_end_tick(), 1440)
        self.assertEqual(clip1.get_audio_clip_data().audio_file_path, TEST_AUDIO_CLIP_DATA["audio_file_path"])
        self.assertEqual(clip1.get_audio_clip_data().start_tick, TEST_AUDIO_CLIP_DATA["start_tick"])
        self.assertAlmostEqual(clip1.get_audio_clip_data().duration, TEST_AUDIO_CLIP_DATA["duration"])

    def test_set_and_get_custom_clip_start_and_end_tick_within_audio_range(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=960,
            clip_end_tick=1000,
            audio_clip_data=TEST_AUDIO_CLIP_DATA,
        )
        self.assertEqual(clip1.get_clip_start_tick(), 960)
        self.assertEqual(clip1.get_clip_end_tick(), 1000)
        self.assertEqual(clip1.get_audio_duration(), 1)
        self.assertEqual(clip1.get_audio_end_tick(), 1440)
        self.assertEqual(clip1.get_audio_clip_data().audio_file_path, TEST_AUDIO_CLIP_DATA["audio_file_path"])
        self.assertEqual(clip1.get_audio_clip_data().start_tick, TEST_AUDIO_CLIP_DATA["start_tick"])
        self.assertAlmostEqual(clip1.get_audio_clip_data().duration, TEST_AUDIO_CLIP_DATA["duration"])
    
    def test_get_optional_clip_audio_data_fields(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=960,
            clip_end_tick=1440,
            audio_clip_data={
                "start_tick": 480,
                "audio_file_path": 'file1',
                "duration": 1,
                "speed_ratio": 2,
                "pitch_offset": 12
            },
        )
        self.assertEqual(clip1.get_audio_speed_ratio(), 2)
        self.assertEqual(clip1.get_audio_pitch_offset(), 12)
        clip1.set_audio_pitch_offset(-13)
        self.assertEqual(clip1.get_audio_pitch_offset(), -13)

    def test_set_and_get_audio_pitch_offset(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=960,
            clip_end_tick=1000,
            audio_clip_data=TEST_AUDIO_CLIP_DATA,
        )
        self.assertEqual(clip1.get_audio_pitch_offset(), 0)
        clip1.set_audio_pitch_offset(24)
        self.assertEqual(clip1.get_audio_pitch_offset(), 24)
        with pytest.raises(Exception) as e_info_1:
            clip1.set_audio_pitch_offset(-99)

        with pytest.raises(Exception) as e_info_2:
            clip1.set_audio_pitch_offset(99)

    def rejects_out_of_range_audio_speed_ratio(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=960,
            clip_end_tick=1440,
            audio_clip_data={
                "start_tick": 480,
                "audio_file_path": 'file1',
                "duration": 1
            },
        )
        with pytest.raises(Exception) as e_info_1:
            clip1.time_stretch_from_clip_right(9999999)
        with pytest.raises(Exception) as e_info_2:
            clip1.time_stretch_from_clip_right(961)
        with pytest.raises(Exception) as e_info_3:
            clip1.time_stretch_from_clip_left(1439)
        with pytest.raises(Exception) as e_info_4:
            clip1.time_stretch_from_clip_left(-999999)

    def test_time_stretch_from_clip_left(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=960,
            clip_end_tick=1440,
            audio_clip_data={
                "start_tick": 480,
                "audio_file_path": 'file1',
                "duration": 1,
            },
        )
        self.assertEqual(clip1.get_raw_audio_duration(), 1)
        self.assertEqual(clip1.get_audio_duration(), 1)
        self.assertEqual(clip1.get_audio_start_tick(), 480)
        self.assertEqual(clip1.get_audio_end_tick(), 1440)
        self.assertEqual(clip1.get_clip_start_tick(), 960)
        self.assertEqual(clip1.get_clip_end_tick(), 1440)
        self.assertEqual(clip1.get_audio_speed_ratio(), 1)

        clip1.time_stretch_from_clip_left(480)
        self.assertEqual(clip1.get_raw_audio_duration(), 1)
        self.assertAlmostEqual(clip1.get_audio_duration(), 2.0)
        self.assertEqual(clip1.get_audio_start_tick(), -480)
        self.assertEqual(clip1.get_audio_end_tick(), 1440)
        self.assertEqual(clip1.get_clip_start_tick(), 480)
        self.assertEqual(clip1.get_clip_end_tick(), 1440)
        self.assertAlmostEqual(clip1.get_audio_speed_ratio(), 0.5)

        clip1.time_stretch_from_clip_left(1200)
        self.assertEqual(clip1.get_raw_audio_duration(), 1)
        self.assertEqual(clip1.get_audio_duration(), 0.5)
        self.assertEqual(clip1.get_audio_start_tick(), 960)
        self.assertEqual(clip1.get_audio_end_tick(), 1440)
        self.assertEqual(clip1.get_clip_start_tick(), 1200)
        self.assertEqual(clip1.get_clip_end_tick(), 1440)
        self.assertAlmostEqual(clip1.get_audio_speed_ratio(), 2)

    def test_time_stretch_from_clip_right(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=960,
            clip_end_tick=1440,
            audio_clip_data={
                "start_tick": 480,
                "audio_file_path": 'file1',
                "duration": 1,
            },
        )
        self.assertEqual(clip1.get_raw_audio_duration(), 1)
        self.assertEqual(clip1.get_audio_duration(), 1)
        self.assertEqual(clip1.get_audio_start_tick(), 480)
        self.assertEqual(clip1.get_audio_end_tick(), 1440)
        self.assertEqual(clip1.get_clip_start_tick(), 960)
        self.assertEqual(clip1.get_clip_end_tick(), 1440)
        self.assertEqual(clip1.get_audio_speed_ratio(), 1)

        clip1.time_stretch_from_clip_right(1200)
        self.assertEqual(clip1.get_audio_duration(), 0.5)
        self.assertEqual(clip1.get_audio_start_tick(), 720)
        self.assertEqual(clip1.get_audio_end_tick(), 1200)
        self.assertEqual(clip1.get_clip_start_tick(), 960)
        self.assertEqual(clip1.get_clip_end_tick(), 1200)
        self.assertAlmostEqual(clip1.get_audio_speed_ratio(), 2.0)

        clip1.time_stretch_from_clip_right(1680)
        self.assertEqual(clip1.get_raw_audio_duration(), 1)
        self.assertEqual(clip1.get_audio_duration(), 2)
        self.assertEqual(clip1.get_audio_start_tick(), 0)
        self.assertEqual(clip1.get_audio_end_tick(), 1680)
        self.assertEqual(clip1.get_clip_start_tick(), 960)
        self.assertEqual(clip1.get_clip_end_tick(), 1680)
        self.assertAlmostEqual(clip1.get_audio_speed_ratio(), 0.5)


class TestTrimLeftAndTrimRight(BaseTestCase):
    def test_trim_left_within_audio_range(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=960,
            clip_end_tick=1000,
            audio_clip_data=TEST_AUDIO_CLIP_DATA,
        )
        self.assertEqual(clip1.get_clip_start_tick(), 960)
        self.assertEqual(clip1.get_clip_end_tick(), 1000)
        clip1.adjust_clip_left(600)
        self.assertEqual(clip1.get_clip_start_tick(), 600)
        self.assertEqual(clip1.get_clip_end_tick(), 1000)

    def test_trim_left_out_of_audio_range(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=960,
            clip_end_tick=1000,
            audio_clip_data=TEST_AUDIO_CLIP_DATA,
        )
        self.assertEqual(clip1.get_clip_start_tick(), 960)
        self.assertEqual(clip1.get_clip_end_tick(), 1000)
        clip1.adjust_clip_left(-100)
        self.assertEqual(clip1.get_clip_start_tick(), 480)
        self.assertEqual(clip1.get_clip_end_tick(), 1000)

    def test_trim_right_within_audio_range(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=960,
            clip_end_tick=1000,
            audio_clip_data=TEST_AUDIO_CLIP_DATA,
        )
        self.assertEqual(clip1.get_clip_start_tick(), 960)
        self.assertEqual(clip1.get_clip_end_tick(), 1000)
        clip1.adjust_clip_right(980)
        self.assertEqual(clip1.get_clip_start_tick(), 960)
        self.assertEqual(clip1.get_clip_end_tick(), 980)

    def test_trim_right_out_of_audio_range(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=960,
            clip_end_tick=1000,
            audio_clip_data=TEST_AUDIO_CLIP_DATA,
        )
        self.assertEqual(clip1.get_clip_start_tick(), 960)
        self.assertEqual(clip1.get_clip_end_tick(), 1000)
        clip1.adjust_clip_right(2000)
        self.assertEqual(clip1.get_clip_start_tick(), 960)
        self.assertEqual(clip1.get_clip_end_tick(), 1440)


class TestMoveClip(BaseTestCase):
    def test_move_clip_within_tempo_range(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=960,
            clip_end_tick=1200,
            audio_clip_data={
                "audio_file_path": 'file1',
                "start_tick": 480,
                "duration": 1,
            },
        )
        self.assertEqual(clip1.get_clip_start_tick(), 960)
        self.assertEqual(clip1.get_clip_end_tick(), 1200)
        self.assertAlmostEqual(clip1.get_duration(), 0.25)
        self.assertEqual(clip1.get_audio_duration(), 1)
        self.assertEqual(clip1.get_audio_end_tick(), 1440)
        clip1.move_clip(-480, move_associated_track_automation_points=False)
        self.assertEqual(clip1.get_clip_start_tick(), 480)
        self.assertEqual(clip1.get_clip_end_tick(), 720)
        self.assertAlmostEqual(clip1.get_duration(), 0.25)
        self.assertEqual(clip1.get_audio_clip_data().audio_file_path, 'file1')
        self.assertEqual(clip1.get_audio_clip_data().duration, 1)
        self.assertEqual(clip1.get_audio_clip_data().start_tick, 0)
        self.assertEqual(clip1.get_audio_end_tick(), 960)

    def test_move_clip_to_the_right_cross_tempo_ranges(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=960,
            clip_end_tick=1200,
            audio_clip_data={
                "audio_file_path": 'file1',
                "start_tick": 480,
                "duration": 1,
            },
        )
        self.assertEqual(clip1.get_clip_start_tick(), 960)
        self.assertEqual(clip1.get_clip_end_tick(), 1200)
        self.assertAlmostEqual(clip1.get_duration(), 0.25)
        self.assertEqual(clip1.get_audio_duration(), 1)
        self.assertEqual(clip1.get_audio_end_tick(), 1440)
        clip1.move_clip(480, move_associated_track_automation_points=False)
        self.assertEqual(clip1.get_clip_start_tick(), 1440)
        self.assertEqual(clip1.get_clip_end_tick(), 1560)
        self.assertAlmostEqual(clip1.get_duration(), 0.25)
        self.assertEqual(clip1.get_audio_clip_data().audio_file_path, 'file1')
        self.assertEqual(clip1.get_audio_clip_data().duration, 1)
        self.assertEqual(clip1.get_audio_clip_data().start_tick, 960)
        self.assertEqual(clip1.get_audio_end_tick(), 1680)

    def test_move_clip_to_the_left_cross_tempo_ranges(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=1320,
            clip_end_tick=1560,
            audio_clip_data={
                "audio_file_path": 'file2',
                "start_tick": 1200,
                "duration": 1,
            },
        )
        self.assertEqual(clip1.get_audio_duration(), 1)
        self.assertEqual(clip1.get_audio_end_tick(), 1800)
        self.assertEqual(clip1.get_clip_start_tick(), 1320)
        self.assertEqual(clip1.get_clip_end_tick(), 1560)
        self.assertAlmostEqual(clip1.get_duration(), 0.375)
        clip1.move_clip(-120, move_associated_track_automation_points=False)
        self.assertEqual(clip1.get_clip_start_tick(), 1200)
        self.assertEqual(clip1.get_clip_end_tick(), 1500)
        self.assertAlmostEqual(clip1.get_duration(), 0.375)
        self.assertEqual(clip1.get_audio_clip_data().audio_file_path, 'file2')
        self.assertEqual(clip1.get_audio_clip_data().duration, 1)
        self.assertEqual(clip1.get_audio_clip_data().start_tick, 1080)
        self.assertEqual(clip1.get_audio_end_tick(), 1740)

    def test_move_clip_to_the_left_cross_0(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=1320,
            clip_end_tick=1560,
            audio_clip_data={
                "audio_file_path": 'file2',
                "start_tick": 1200,
                "duration": 1,
            },
        )
        self.assertEqual(clip1.get_audio_end_tick(), 1800)
        self.assertEqual(clip1.get_clip_start_tick(), 1320)
        self.assertEqual(clip1.get_clip_end_tick(), 1560)
        self.assertAlmostEqual(clip1.get_duration(), 0.375)
        clip1.move_clip(-1440, move_associated_track_automation_points=False)
        self.assertEqual(clip1.get_clip_start_tick(), 0)
        self.assertEqual(clip1.get_clip_end_tick(), 240)
        self.assertAlmostEqual(clip1.get_duration(), 0.25)
        self.assertEqual(clip1.get_audio_clip_data().audio_file_path, 'file2')
        self.assertEqual(clip1.get_audio_clip_data().duration, 1)
        self.assertEqual(clip1.get_audio_clip_data().start_tick, -240)
        self.assertEqual(clip1.get_audio_end_tick(), 720)

    def test_move_clip_to_the_left_cross_tempo_and_0(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=1920,
            clip_end_tick=2040,
            audio_clip_data={
                "audio_file_path": 'file3',
                "start_tick": 1680,
                "duration": 1,
            },
        )
        self.assertEqual(clip1.get_audio_end_tick(), 2160)
        self.assertEqual(clip1.get_clip_start_tick(), 1920)
        self.assertEqual(clip1.get_clip_end_tick(), 2040)
        self.assertAlmostEqual(clip1.get_duration(), 0.25)
        clip1.move_clip(-2040, move_associated_track_automation_points=False)
        self.assertEqual(clip1.get_clip_start_tick(), 0)
        self.assertEqual(clip1.get_clip_end_tick(), 120)
        self.assertAlmostEqual(clip1.get_duration(), 0.125)
        self.assertEqual(clip1.get_audio_clip_data().audio_file_path, 'file3')
        self.assertEqual(clip1.get_audio_clip_data().duration, 1)
        self.assertEqual(clip1.get_audio_clip_data().start_tick, -600)
        self.assertEqual(clip1.get_audio_end_tick(), 360)


class TestResolveConflict(BaseTestCase):
    def test_create_clip_in_the_middle_of_another_clip(self):
        self.audio_track.create_audio_clip(
            clip_start_tick=0,
            clip_end_tick=960,
            audio_clip_data={
                "audio_file_path": 'file1',
                "start_tick": 0,
                "duration": 1,
            },
        )
        self.audio_track.create_audio_clip(
            clip_start_tick=240,
            clip_end_tick=720,
            audio_clip_data={
                "audio_file_path": 'file2',
                "start_tick": 240,
                "duration": 0.5,
            },
        )
        self.assertEqual(self.audio_track.get_clip_count(), 3)
        self.assertEqual(self.audio_track.get_clip_at(0).get_clip_start_tick(), 0)
        self.assertEqual(self.audio_track.get_clip_at(0).get_clip_end_tick(), 239)
        self.assertEqual(self.audio_track.get_clip_at(0).get_audio_clip_data().audio_file_path, 'file1')
        self.assertEqual(self.audio_track.get_clip_at(0).get_audio_clip_data().duration, 1)
        self.assertEqual(self.audio_track.get_clip_at(0).get_audio_clip_data().start_tick, 0)
        self.assertEqual(self.audio_track.get_clip_at(1).get_clip_start_tick(), 240)
        self.assertEqual(self.audio_track.get_clip_at(1).get_clip_end_tick(), 720)
        self.assertEqual(self.audio_track.get_clip_at(1).get_audio_clip_data().audio_file_path, 'file2')
        self.assertAlmostEqual(self.audio_track.get_clip_at(1).get_audio_clip_data().duration, 0.5)
        self.assertEqual(self.audio_track.get_clip_at(1).get_audio_clip_data().start_tick, 240)
        self.assertEqual(self.audio_track.get_clip_at(2).get_clip_start_tick(), 721)
        self.assertEqual(self.audio_track.get_clip_at(2).get_clip_end_tick(), 960)
        self.assertEqual(self.audio_track.get_clip_at(2).get_audio_clip_data().audio_file_path, 'file1')
        self.assertAlmostEqual(self.audio_track.get_clip_at(2).get_audio_clip_data().duration, 1)
        self.assertEqual(self.audio_track.get_clip_at(2).get_audio_clip_data().start_tick, 0)

    def test_create_clip_to_the_left_of_another_clip(self):
        self.audio_track.create_audio_clip(
            clip_start_tick=0,
            clip_end_tick=960,
            audio_clip_data={
                "audio_file_path": 'file1',
                "start_tick": 0,
                "duration": 1,
            },
        )
        self.audio_track.create_audio_clip(
            clip_start_tick=0,
            clip_end_tick=240,
            audio_clip_data={
                "audio_file_path": 'file2',
                "start_tick": 0,
                "duration": 0.25,
            },
        )
        self.assertEqual(self.audio_track.get_clip_count(), 2)
        self.assertEqual(self.audio_track.get_clip_at(0).get_clip_start_tick(), 0)
        self.assertEqual(self.audio_track.get_clip_at(0).get_clip_end_tick(), 240)
        self.assertEqual(self.audio_track.get_clip_at(0).get_audio_clip_data().audio_file_path, 'file2')
        self.assertAlmostEqual(self.audio_track.get_clip_at(0).get_audio_clip_data().duration, 0.25)
        self.assertEqual(self.audio_track.get_clip_at(0).get_audio_clip_data().start_tick, 0)
        self.assertEqual(self.audio_track.get_clip_at(1).get_clip_start_tick(), 241)
        self.assertEqual(self.audio_track.get_clip_at(1).get_clip_end_tick(), 960)
        self.assertEqual(self.audio_track.get_clip_at(1).get_audio_clip_data().audio_file_path, 'file1')
        self.assertAlmostEqual(self.audio_track.get_clip_at(1).get_audio_clip_data().duration, 1)
        self.assertEqual(self.audio_track.get_clip_at(1).get_audio_clip_data().start_tick, 0)

    def test_create_clip_to_the_right_of_another_clip(self):
        self.audio_track.create_audio_clip(
            clip_start_tick=0,
            clip_end_tick=960,
            audio_clip_data={
                "audio_file_path": 'file1',
                "start_tick": 0,
                "duration": 1,
            },
        )
        self.audio_track.create_audio_clip(
            clip_start_tick=720,
            clip_end_tick=960,
            audio_clip_data={
                "audio_file_path": 'file2',
                "start_tick": 720,
                "duration": 0.25,
            },
        )
        self.assertEqual(self.audio_track.get_clip_count(), 2)
        self.assertEqual(self.audio_track.get_clip_at(0).get_clip_start_tick(), 0)
        self.assertEqual(self.audio_track.get_clip_at(0).get_clip_end_tick(), 719)
        self.assertEqual(self.audio_track.get_clip_at(0).get_audio_clip_data().audio_file_path, 'file1')
        self.assertAlmostEqual(self.audio_track.get_clip_at(0).get_audio_clip_data().duration, 1)
        self.assertEqual(self.audio_track.get_clip_at(0).get_audio_clip_data().start_tick, 0)

        self.assertEqual(self.audio_track.get_clip_at(1).get_clip_start_tick(), 720)
        self.assertEqual(self.audio_track.get_clip_at(1).get_clip_end_tick(), 960)
        self.assertEqual(self.audio_track.get_clip_at(1).get_audio_clip_data().audio_file_path, 'file2')
        self.assertAlmostEqual(self.audio_track.get_clip_at(1).get_audio_clip_data().duration, 0.25)
        self.assertEqual(self.audio_track.get_clip_at(1).get_audio_clip_data().start_tick, 720)

    def test_trim_left_edge_overlaps_with_another_clip(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=0,
            clip_end_tick=480,
            audio_clip_data={
                "audio_file_path": 'file1',
                "start_tick": 0,
                "duration": 0.5,
            },
        )
        clip2 = self.audio_track.create_audio_clip(
            clip_start_tick=720,
            clip_end_tick=960,
            audio_clip_data={
                "audio_file_path": 'file2',
                "start_tick": 360,
                "duration": 1,
            },
        )
        self.assertEqual(self.audio_track.get_clip_count(), 2)
        clip2.adjust_clip_left(240, resolve_conflict=True)
        self.assertEqual(clip1.get_clip_start_tick(), 0)
        self.assertEqual(clip1.get_clip_end_tick(), 359)
        self.assertEqual(clip1.get_audio_clip_data().audio_file_path, 'file1')
        self.assertAlmostEqual(clip1.get_audio_clip_data().duration, 0.5)
        self.assertEqual(clip1.get_audio_clip_data().start_tick, 0)
        self.assertEqual(clip2.get_clip_start_tick(), 360)
        self.assertEqual(clip2.get_clip_end_tick(), 960)
        self.assertEqual(clip2.get_audio_clip_data().audio_file_path, 'file2')
        self.assertAlmostEqual(clip2.get_audio_clip_data().duration, 1)
        self.assertEqual(clip2.get_audio_clip_data().start_tick, 360)

    def test_trim_right_edge_overlaps_with_another_clip(self):
        clip1 = self.audio_track.create_audio_clip(
            clip_start_tick=0,
            clip_end_tick=240,
            audio_clip_data={
                "audio_file_path": 'file1',
                "start_tick": 0,
                "duration": 0.75,
            },
        )
        clip2 = self.audio_track.create_audio_clip(
            clip_start_tick=480,
            clip_end_tick=960,
            audio_clip_data={
                "audio_file_path": 'file2',
                "start_tick": 360,
                "duration": 1,
            },
        )
        self.assertEqual(self.audio_track.get_clip_count(), 2)
        clip1.adjust_clip_right(800, resolve_conflict=True)
        self.assertEqual(clip1.get_clip_start_tick(), 0)
        self.assertEqual(clip1.get_clip_end_tick(), 720)
        self.assertEqual(clip1.get_audio_clip_data().audio_file_path, 'file1')
        self.assertAlmostEqual(clip1.get_audio_clip_data().duration, 0.75)
        self.assertEqual(clip1.get_audio_clip_data().start_tick, 0)
        self.assertEqual(clip2.get_clip_start_tick(), 721)
        self.assertEqual(clip2.get_clip_end_tick(), 960)
        self.assertEqual(clip2.get_audio_clip_data().audio_file_path, 'file2')
        self.assertAlmostEqual(clip2.get_audio_clip_data().duration, 1)
        self.assertEqual(clip2.get_audio_clip_data().start_tick, 360)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from tuneflow_py.descriptors.clip_descriptor import AudioClipData
from tuneflow_py.models.protos import song_pb2
from tuneflow_py.models.clip import Clip, ClipType
from tuneflow_py.models.note import Note
from tuneflow_py.models.audio_plugin import AudioPlugin
from tuneflow_py.models.automation import AutomationData
from tuneflow_py.utils import db_to_volume_value, volume_value_to_db, lower_equal, greater_equal, lower_than, decode_audio_plugin_tuneflow_id
import nanoid
from typing import List
from types import SimpleNamespace

TrackType = song_pb2.TrackType
TrackOutputType = song_pb2.TrackOutput.TrackOutputType

class TrackOutput:
    def __init__(self, proto: song_pb2.TrackOutput | None  =None) -> None:
        if not proto:
            self._proto = song_pb2.TrackOutput()
        else:
            self._proto = proto
    
    def get_type(self):
        return self._proto.type
    
    def set_type(self, type: int):
        self._proto.type = type

    def get_track_id(self):
        return self._proto.track_id

    def set_track_id(self, track_id: str):
        self._proto.track_id = track_id

class Track:
    def __init__(self, type: int | None = None,
                 song=None,
                 uuid=None,
                 instrument=None,
                 volume=db_to_volume_value(0),
                 solo=False,
                 muted=False,
                 rank=0,
                 pan=0, proto: song_pb2.Track | None = None) -> None:
        '''
        A track in the song that maps to an instrument.

        It contains clips, instrument information, play status(volume, muted, etc.), and more.

        IMPORTANT: Do not use the constructor directly, call create_track from a song instead.

        @param uuid The universal-unique identifier of the track. In most cases, leave it blank and it will be automatically assigned.
        @param instrument Information about the instrument to play this track.
        @param volume A float value indicating the track-level volume, ranging from 0 to 1.
        @param pan An integer value from -64 to 63, corresponding to the midi pan CC 0 - 127.
        @param solo Whether this track is in solo mode.
        @param muted Whether this track is muted.
        @param rank The rank of this track within the song.
        '''
        if song is None:
            raise Exception('song must be provided when creating a track')
        self.song = song
        if proto is not None:
            self._proto = proto
            return
        self._proto = song_pb2.Track()
        self._proto.type = type
        if instrument is not None:
            self._proto.instrument.program = instrument["program"]
            self._proto.instrument.is_drum = instrument["is_drum"]
        elif type == TrackType.MIDI_TRACK:
            self._proto.instrument.program = 0
            self._proto.instrument.is_drum = False
        if type == TrackType.AUX_TRACK:
            # TODO: Initialize aux track data.
            pass
        self._proto.uuid = uuid if uuid is not None else Track._generate_track_id()
        self._proto.volume = volume
        self._proto.solo = solo
        self._proto.muted = muted
        self._proto.rank = rank
        self._proto.pan = pan

    def get_id(self):
        return self._proto.uuid

    def get_clip_count(self):
        return len(self._proto.clips)

    def get_clips(self):
        for i in range(len(self._proto.clips)):
            yield self.get_clip_at(i)

    def get_clip_at(self, clip_index: int):
        return self._create_clip_from_proto(proto=self._proto.clips[clip_index])

    def get_type(self) -> int:
        return self._proto.type

    def get_volume(self):
        '''
        Returns the fader position, ranging from 0 to 1.
        '''
        return self._proto.volume

    def get_volume_in_db(self):
        return volume_value_to_db(self.get_volume())

    def set_volume(self, volume: float):
        '''
        Sets the volume fader position.

        @param volume The track volume fader position, ranging from 0 to 1.
        '''
        self._proto.volume = volume

    def get_pan(self) -> int:
        '''
        A pan value from -64 to 63
        '''
        return self._proto.pan

    def set_pan(self, pan: int):
        '''
        Sets the track pan knob position.

        @param pan An integer value between -64 and 63. Setting to 0 means balanced.
        '''
        self._proto.pan = pan

    def get_solo(self) -> bool:
        '''
        Returns whether the track is being solo'ed.
        '''
        return self._proto.solo

    def set_solo(self, solo: bool):
        '''
        @param solo If set to true, track will be solo'ed and unmuted.
        '''
        self._proto.solo = solo

    def get_muted(self) -> bool:
        return self._proto.muted

    def set_muted(self, muted: bool):
        self._proto.muted = muted

    def get_rank(self):
        return self._proto.rank

    def has_instrument(self):
        return self._proto.HasField('instrument')

    def get_instrument(self):
        if not self.has_instrument():
            return None
        return self._proto.instrument

    def set_instrument(self, program: int, is_drum: bool):
        '''
        Set the midi instrument that this track corresponds to. This will also
        be the instrument to be used by the Minimal playback engine.

        @param program General MIDI program number(counting from 0, i.e. "Acoustic Grand Piano" === 0). 
            https://www.midi.org/specifications-old/item/gm-level-1-sound-set

        @param is_drum Whether this instrument is a drumset.
            '''
        if self.get_type() != TrackType.MIDI_TRACK:
            return
        self._proto.instrument.program = program
        self._proto.instrument.is_drum = is_drum

    def has_sampler_plugin(self):
        return self._proto.HasField('sampler_plugin')

    def get_sampler_plugin(self):
        '''
        Gets the synth/sampler plugin of this track.

        The sampler plugin generates sound from MIDI notes and will only work for MIDI tracks.
        '''
        if not self.has_sampler_plugin():
            return None
        return AudioPlugin(proto=self._proto.sampler_plugin)

    def set_sampler_plugin(self, plugin: AudioPlugin | None, clear_automation=True):
        '''
        Sets the synth/sampler plugin of this track.

        @param clear_automation Whether to remove existing track automation associated with the old plugin.
        '''
        if (self.get_type() != TrackType.MIDI_TRACK):
            return

        old_plugin = self.get_sampler_plugin()
        if plugin is not None:
            self._proto.sampler_plugin.MergeFrom(plugin._proto)
        else:
            self._proto.ClearField("sampler_plugin")
        if clear_automation:
            plugin_type_changed = (
                self.has_sampler_plugin() == False and plugin is not None) or (
                plugin is None and self.has_sampler_plugin()) or (
                plugin is not None and self.has_sampler_plugin() and not plugin.matches_tf_id(
                    self.get_sampler_plugin().get_tuneflow_id()))  # type:ignore
            if (plugin_type_changed and old_plugin is not None):
                self.get_automation().remove_automation_of_plugin(old_plugin.get_instance_id())

    def get_audio_plugin_count(self):
        return len(self._proto.audio_plugin)

    def get_audio_plugin_at(self, index):
        return AudioPlugin(proto=self._proto.audio_plugin[index])

    def get_suggested_instruments_count(self):
        return len(self._proto.suggested_instruments)

    def get_suggested_instrument_at(self, index: int):
        self._proto.suggested_instruments[index]

    def create_suggested_instrument(self, program: int, is_drum: bool):
        '''
        Adds a suggested instrument and returns it.

        @param program General MIDI program number(counting from 0, i.e. "Acoustic Grand Piano" === 0). 
            https://www.midi.org/specifications-old/item/gm-level-1-sound-set

        @param is_drum Whether this instrument is a percussion instrument
            (or using channel 9(counting from 0) if you know what it means).
        '''
        if (self.get_type() != TrackType.MIDI_TRACK):
            return
        instrument_info = self._proto.suggested_instruments.add(
            program=program, is_drum=is_drum)
        return instrument_info

    def clear_suggested_instruments(self):
        del self._proto.suggested_instruments[:]

    def get_track_start_tick(self):
        if self.get_clip_count() == 0:
            return 0
        return self.get_clip_at(0).get_clip_start_tick()

    def get_track_end_tick(self):
        if self.get_clip_count() == 0:
            return 0
        return self.get_clip_at(self.get_clip_count() - 1).get_clip_end_tick()

    def get_clip_by_id(self, clip_id: str):
        for clip_proto in self._proto.clips:
            if clip_proto.id == clip_id:
                return self._create_clip_from_proto(proto=clip_proto)
        return None

    def create_midi_clip(self, clip_start_tick: int, clip_end_tick: int | None = None, insert_clip=True):
        '''
        Creates a MIDI clip and optionally inserts it into the track.

        @param clip_start_tick The start of the clip, must be specified.
        @param insert_clip Whether to insert the created clip into the track.
        '''
        if clip_start_tick is None:
            raise Exception(
                'clip_start_tick must be specified when creating a clip.')

        new_clip_end_tick = clip_start_tick + \
            1 if clip_end_tick is None else clip_end_tick
        if new_clip_end_tick < clip_start_tick:
            raise Exception(
                f'clip_end_tick must be greater or equal to clip_start_tick, got clip_start_tick: {clip_start_tick}, clip_end_tick: {clip_end_tick}'
            )

        clip = Clip(
            id=Clip._generate_clip_id(),
            type=ClipType.MIDI_CLIP,
            song=self.song,
            track=None,
            clip_start_tick=clip_start_tick,
            clip_end_tick=new_clip_end_tick,
        )
        if insert_clip:
            self.insert_clip(clip)

        return clip

    def create_audio_clip(
            self, clip_start_tick: int, audio_clip_data: AudioClipData, clip_end_tick: int | None = None, insert_clip=True):
        '''
        Creates an audio clip and optionally inserts it into the track.

        @param clip_start_tick The start of the clip, must be specified.
        @param audio_clip_data A dict of audio-related data.
        @param insert_clip Whether to insert the created clip into the track.
        '''
        if clip_start_tick is None:
            raise Exception(
                'clip_start_tick must be specified when creating a clip.')

        clip = Clip(
            id=Clip._generate_clip_id(),
            type=ClipType.AUDIO_CLIP,
            song=self.song,
            track=None,
            clip_start_tick=clip_start_tick,
            clip_end_tick=clip_end_tick,
            audio_clip_data=audio_clip_data,
        )
        if (insert_clip):
            self.insert_clip(clip)

        return clip

    def insert_clip(self, clip: Clip):
        if clip.get_track() != self:
            if clip.get_track() is not None:
                # Clip belongs to another track.
                clip.delete_from_parent(
                    delete_associated_track_automation=False)
            clip.track = self
        else:
            # Clip already belongs to the track.
            return

        # Resolve conflict before inserting a new clip
        # to preserve the current order of clips.
        self._resolve_clip_conflict(
            clip.get_id(), clip.get_clip_start_tick(), clip.get_clip_end_tick())
        self._ordered_insert_clip(clip)

    def get_clip_index(self, clip: Clip):
        '''
        Get the index of the clip within the clip list.

        NOTE: This assumes the clip list is sorted.
        '''
        start_index = lower_equal(
            self._proto.clips,
            clip._proto,
            key=lambda x: x.clip_start_tick,
        )

        for i in range(max(0, start_index), len(self._proto.clips)):
            if self._proto.clips[i].id == clip.get_id():
                return i
        return -1

    def delete_clip(self, clip: Clip, delete_associated_track_automation: bool):
        index = self.get_clip_index(clip)
        if index >= 0 and index < self.get_clip_count():
            self.delete_clip_at(index, delete_associated_track_automation)
            clip.track = None

    def delete_clip_at(self, index: int, delete_associated_track_automation: bool):
        if (index < 0):
            return

        if delete_associated_track_automation:
            if index < 0 or index >= len(self._proto.clips):
                return

            # Remove automation points within range
            clip = self.get_clip_at(index)
            self.get_automation().remove_all_points_within_range(clip.get_clip_start_tick(), clip.get_clip_end_tick())

        self._proto.clips.pop(index)

    def get_clips_overlapping_with(self, start_tick: int, end_tick: int):
        '''
        Gets the clips whose range overlaps with the given range.
        '''
        overlapping_clips: List[Clip] = []
        target_clip = SimpleNamespace()
        target_clip.clip_start_tick = start_tick
        start_index = lower_than(
            self._proto.clips,
            target_clip,
            key=lambda x: x.clip_start_tick,
        )
        for i in range(max(start_index, 0), len(self._proto.clips)):
            current_clip_proto = self._proto.clips[i]
            if (current_clip_proto.clip_end_tick < start_tick):
                continue

            if current_clip_proto.clip_start_tick > end_tick:
                break

            overlapping_clips.append(self._create_clip_from_proto(proto=current_clip_proto))

        return overlapping_clips

    def create_audio_plugin(self, tf_id: str):
        plugin_info = decode_audio_plugin_tuneflow_id(tf_id)
        plugin = AudioPlugin(
            name=plugin_info["name"],
            manufacturer_name=plugin_info["manufacturer_name"],
            plugin_format_name=plugin_info["plugin_format_name"],
            plugin_version=plugin_info["plugin_version"],
        )
        return plugin

    def get_automation(self):
        return AutomationData(self._proto.automation)

    def clone_clip(self, clip: Clip):
        '''
        Clones a clip without inserting it into this track, and returns the cloned instance.

        @param clip The clip (not necessarily in this track) to clone.
        @returns The cloned clip.
        '''
        new_clip_proto = song_pb2.Clip()
        new_clip_proto.CopyFrom(clip._proto)
        new_clip_proto.id = Clip._generate_clip_id()
        return Clip(proto=new_clip_proto, song=self.song)

    def has_output(self):
        return self._proto.HasField('output')
    
    def get_output(self):
        if not self.has_output():
            return None
        return TrackOutput(proto=self._proto.output)

    def get_or_create_output(self):
        return TrackOutput(proto=self._proto.output)
    
    def remove_output(self):
        self._proto.ClearField('output')

    def _resolve_clip_conflict(self, clip_id: str, start_tick: int, end_tick: int):
        overlapping_clips = self.get_clips_overlapping_with(
            start_tick, end_tick)
        for clip in overlapping_clips:
            if clip.get_id() == clip_id:
                continue
            clip._trim_conflict_part(start_tick, end_tick)

    def _ordered_insert_clip(self, new_clip: Clip):
        insert_index = greater_equal(
            self._proto.clips,
            new_clip._proto,
            key=lambda x: x.clip_start_tick,
        )

        self._proto.clips.insert(insert_index, new_clip._proto)
        # Re-assign proto since protobuf created a new copy for the inserted proto
        new_clip._proto = self._proto.clips[insert_index]

    def _create_clip_from_proto(self, proto: song_pb2.Clip):
        return Clip(song=self.song, track=self, proto=proto)

    def __repr__(self) -> str:
        return str(self._proto)

    @staticmethod
    def _generate_track_id():
        return nanoid.generate()

    def get_visible_notes(self) -> List[Note]:
        visible_notes = []
        for clip in self.get_clips():
            for note in clip.get_notes():
                visible_notes.append(note)
        return sorted(visible_notes, key=lambda note: note.get_start_tick())



from typing_extensions import TypedDict, Required, NotRequired


class AudioPluginDescriptor(TypedDict):
    '''
    Descriptor of an audio plugin.
    '''

    category: NotRequired[str]
    '''
    Category of the plugin.
    e.g.: 'Instrument|Synth', 'Fx|EQ', etc.
    '''

    descriptiveName: NotRequired[str]
    ''' A more detailed name. '''

    id: Required[str]
    ''' An internal id of the plugin. '''

    isInstrument: Required[str]
    '''
    Whether this plugin can be used as an instrument to generate
    sound from MIDI notes.
    '''

    manufacturerName: Required[str]
    ''' Manufacturer of the plugin. '''

    name: Required[str]
    '''
    A short name of the plugin.
    Use this when creating `AudioPlugin` instances.
    '''

    pluginFormatName: Required[str]
    ''' Format of the plugin. e.g. 'VST3', 'AudioUnit', etc. '''

    version: Required[str]
    ''' Version of the plugin. '''



from typing import Union
from typing_extensions import Literal, TypedDict, Required, NotRequired

RealNumber = Union[int, float]


class ClipInfo(TypedDict):
    track_id: Required[str]
    clip_id: Required[str]


class EntityId(TypedDict):
    type: Required[Literal['song', 'track', 'clip', 'note']]
    trackId: NotRequired[str]
    clipId: NotRequired[str]
    noteId: NotRequired[int]


SupportedPlatform = Literal['desktop', 'web']



import unittest
from unittest.mock import ANY
from tuneflow_py import Song, TrackType, AutomationTarget, AutomationTargetType, AutomationValue, AutomationData, AutomationPoint


def create_song():
    song = Song()
    midi_track = song.create_track(type=TrackType.MIDI_TRACK)
    return song, midi_track


class AnyInt(int):
    def __eq__(self, __value: object) -> bool:
        return True


class CloseFloat(float):
    def __eq__(self, __value: object) -> bool:
        return abs(self - __value) < 0.00001


def points_to_objects(points):
    return [{
        "id": point.id,
        "value": point.value,
        "tick": point.tick
    } for point in points]


class AutomationTargetTestCase(unittest.TestCase):
    def test_gets_and_sets_properties(self):
        target = AutomationTarget(type=AutomationTargetType.VOLUME)
        self.assertEqual(target.get_type(), AutomationTargetType.VOLUME)
        self.assertEqual(target.get_type(), AutomationTargetType.VOLUME)
        target = AutomationTarget(
            type=AutomationTargetType.AUDIO_PLUGIN,
            plugin_instance_id='someplugininstanceid',
            param_id='someparamid',
        )
        self.assertEqual(target.get_type(), AutomationTargetType.AUDIO_PLUGIN)
        self.assertEqual(target.get_plugin_instance_id(), 'someplugininstanceid')
        self.assertEqual(target.get_param_id(), 'someparamid')
        target.set_plugin_instance_id('anotherplugininstanceid')
        self.assertEqual(target.get_plugin_instance_id(), 'anotherplugininstanceid')
        target.set_param_id('anotherparamid')
        self.assertEqual(target.get_param_id(), 'anotherparamid')
        target.set_type(AutomationTargetType.PAN)
        self.assertEqual(target.get_type(), AutomationTargetType.PAN)

    def test_encodes_target(self):
        target1 = AutomationTarget(type=AutomationTargetType.VOLUME)
        self.assertEqual(target1.to_tf_automation_target_id(), '1')
        target2 = AutomationTarget(
            type=AutomationTargetType.AUDIO_PLUGIN,
            plugin_instance_id='pluginId1',
            param_id='paramId1',
        )
        self.assertEqual(target2.to_tf_automation_target_id(), '3^^pluginId1^^paramId1')

    def test_checks_equality(self):
        target1 = AutomationTarget(type=AutomationTargetType.PAN)
        target2 = AutomationTarget(type=AutomationTargetType.PAN)
        target3 = AutomationTarget(type=AutomationTargetType.VOLUME)
        target4 = AutomationTarget(
            AutomationTargetType.AUDIO_PLUGIN,
            'pluginId1',
            'paramId1',
        )
        target5 = AutomationTarget(
            AutomationTargetType.AUDIO_PLUGIN,
            'pluginId2',
            'paramId2',
        )
        target6 = AutomationTarget(
            AutomationTargetType.AUDIO_PLUGIN,
            'pluginId2',
            'paramId2',
        )
        self.assertEqual(target1.equals(target2), True)
        self.assertEqual(target1.equals(target3), False)
        self.assertEqual(target1.equals(target4), False)
        self.assertEqual(target4.equals(target5), False)
        self.assertEqual(target5.equals(target6), True)

    def test_clones_target(self):
        target1 = AutomationTarget(type=AutomationTargetType.VOLUME)
        target2 = target1.clone()
        self.assertEqual(target1 == target2, False)
        self.assertEqual(target1.equals(target2), True)


class AutomationValueTestCase(unittest.TestCase):
    def test_sets_and_gets_disabled(self):
        automation_value = AutomationValue()
        self.assertEqual(automation_value.get_disabled(), False)
        automation_value.set_disabled(True)
        self.assertEqual(automation_value.get_disabled(), True)

    def test_adds_and_gets_points(self):
        automation_value = AutomationValue()
        self.assertEqual([{
            "id": point.id,
            "tick": point.tick,
            "value": point.value
        } for point in automation_value.get_points()], [])
        automation_value.add_point(tick=3, value=0.5)
        automation_value.add_point(tick=1, value=0.25)
        automation_value.add_point(tick=2, value=0.75)
        self.assertEqual(points_to_objects(automation_value.get_points()), [
            {
                "id": 2,
                "tick": 1,
                "value": CloseFloat(0.25),
            },
            {
                "id": 3,
                "tick": 2,
                "value": CloseFloat(0.75),
            },
            {
                "id": 1,
                "tick": 3,
                "value": CloseFloat(0.5),
            },
        ])
        automation_value.add_point(tick=2, value=1)
        self.assertEqual(points_to_objects(automation_value.get_points()), [
            {
                "id": 2,
                "tick": 1,
                "value": CloseFloat(0.25),
            },
            {
                "id": 4,
                "tick": 2,
                "value": CloseFloat(1),
            },
            {
                "id": 3,
                "tick": 2,
                "value": CloseFloat(0.75),
            },
            {
                "id": 1,
                "tick": 3,
                "value": CloseFloat(0.5),
            },
        ])
        automation_value.add_point(tick=2, value=0.1, overwrite=True)
        self.assertEqual(points_to_objects(automation_value.get_points()), [
            {
                "id": 2,
                "tick": 1,
                "value": CloseFloat(0.25),
            },
            {
                "id": 5,
                "tick": 2,
                "value": CloseFloat(0.1),
            },
            {
                "id": 1,
                "tick": 3,
                "value": CloseFloat(0.5),
            },
        ])

    def test_get_points_within_range(self):
        automation_value = AutomationValue()
        self.assertEqual([{
            "id": point.id,
            "tick": point.tick,
            "value": point.value
        } for point in automation_value.get_points()], [])
        automation_value.add_point(tick=3, value=0.5)
        automation_value.add_point(tick=3, value=0.65)
        automation_value.add_point(tick=1, value=0.25)
        automation_value.add_point(tick=2, value=0.75)
        automation_value.add_point(tick=2, value=0.55)
        automation_value.add_point(tick=4, value=0.15)
        self.assertEqual(points_to_objects(automation_value.get_points_in_range(2, 3)), [
            {
                "id": 5,
                "tick": 2,
                "value": CloseFloat(0.55),
            },
            {
                "id": 4,
                "tick": 2,
                "value": CloseFloat(0.75),
            },
            {
                "id": 2,
                "tick": 3,
                "value": CloseFloat(0.65),
            },
            {
                "id": 1,
                "tick": 3,
                "value": CloseFloat(0.5),
            },
        ])

        self.assertEqual(points_to_objects(automation_value.get_points_in_range(4, 3)), [])

        self.assertEqual(points_to_objects(automation_value.get_points_in_range(5, 6)), [])

        self.assertEqual(points_to_objects(automation_value.get_points_in_range(4, 5)), [
            {
                "id": 6,
                "tick": 4,
                "value": CloseFloat(0.15),
            },
        ])
        self.assertEqual(points_to_objects(automation_value.get_points_in_range(0, 1)), [
            {
                "id": 3,
                "tick": 1,
                "value": CloseFloat(0.25),
            },
        ])

    def test_removes_points(self):
        automation_value = AutomationValue()
        self.assertEqual(points_to_objects(automation_value.get_points()), [])
        automation_value.add_point(tick=3, value=0.5)
        automation_value.add_point(tick=1, value=0.25)
        automation_value.add_point(tick=2, value=0.75)
        automation_value.add_point(tick=4, value=0.15)
        self.assertEqual(points_to_objects(automation_value.get_points()), [
            {
                "id": 2,
                "tick": 1,
                "value": CloseFloat(0.25),
            },
            {
                "id": 3,
                "tick": 2,
                "value": CloseFloat(0.75),
            },
            {
                "id": 1,
                "tick": 3,
                "value": CloseFloat(0.5),
            },
            {
                "id": 4,
                "tick": 4,
                "value": CloseFloat(0.15),
            },
        ])
        points = automation_value.get_points()
        point2Id = points[2].id
        automation_value.remove_points([points[0].id, points[1].id, points[3].id])
        self.assertEqual(points_to_objects(automation_value.get_points()), [
            {
                "id": point2Id,
                "tick": 3,
                "value": CloseFloat(0.5),
            },
        ])

    def test_remove_points_in_range(self):
        automation_value = AutomationValue()
        self.assertEqual(points_to_objects(automation_value.get_points()), [])
        automation_value.add_point(tick=3, value=0.5)
        automation_value.add_point(tick=3, value=0.65)
        automation_value.add_point(tick=1, value=0.25)
        automation_value.add_point(tick=2, value=0.75)
        automation_value.add_point(tick=2, value=0.55)
        automation_value.add_point(tick=4, value=0.15)
        self.assertEqual(points_to_objects(automation_value.get_points()), [
            {
                "id": 3,
                "tick": 1,
                "value": CloseFloat(0.25),
            },
            {
                "id": 5,
                "tick": 2,
                "value": CloseFloat(0.55),
            },
            {
                "id": 4,
                "tick": 2,
                "value": CloseFloat(0.75),
            },
            {
                "id": 2,
                "tick": 3,
                "value": CloseFloat(0.65),
            },
            {
                "id": 1,
                "tick": 3,
                "value": CloseFloat(0.5),
            },
            {
                "id": 6,
                "tick": 4,
                "value": CloseFloat(0.15),
            },
        ])
        automation_value.remove_points_in_range(2, 3)
        self.assertEqual(points_to_objects(automation_value.get_points()), [
            {
                "id": 3,
                "tick": 1,
                "value": CloseFloat(0.25),
            },
            {
                "id": 6,
                "tick": 4,
                "value": CloseFloat(0.15),
            },
        ])

    def test_moves_multiple_points_to_left_overwrite(self):
        automation_value = AutomationValue()
        self.assertEqual(points_to_objects(automation_value.get_points()), [])
        automation_value.add_point(tick=3, value=0.5)
        automation_value.add_point(tick=1, value=0.25)
        automation_value.add_point(tick=1, value=0.65)
        automation_value.add_point(tick=1, value=0.95)
        automation_value.add_point(tick=2, value=0.75)
        automation_value.add_point(tick=2, value=0.85)
        automation_value.add_point(tick=2, value=0.35)
        automation_value.add_point(tick=4, value=0.15)
        self.assertEqual(points_to_objects(automation_value.get_points()), [
            {
                "id": 4,
                "tick": 1,
                "value": CloseFloat(0.95),
            },
            {
                "id": 3,
                "tick": 1,
                "value": CloseFloat(0.65),
            },
            {
                "id": 2,
                "tick": 1,
                "value": CloseFloat(0.25),
            },
            {
                "id": 7,
                "tick": 2,
                "value": CloseFloat(0.35),
            },
            {
                "id": 6,
                "tick": 2,
                "value": CloseFloat(0.85),
            },
            {
                "id": 5,
                "tick": 2,
                "value": CloseFloat(0.75),
            },
            {
                "id": 1,
                "tick": 3,
                "value": CloseFloat(0.5),
            },
            {
                "id": 8,
                "tick": 4,
                "value": CloseFloat(0.15),
            },
        ])
        points = automation_value.get_points()
        automation_value.move_points([points[1].id, points[2].id, points[3].id, points[4].id], -1, 0.3)
        self.assertEqual(points_to_objects(automation_value.get_points()), [
            {
                "id": 3,
                "tick": 0,
                "value": CloseFloat(0.95),
            },
            {
                "id": 2,
                "tick": 0,
                "value": CloseFloat(0.55),
            },
            {
                "id": 7,
                "tick": 1,
                "value": CloseFloat(0.65),
            },
            {
                "id": 6,
                "tick": 1,
                "value": 1,
            },
            {
                "id": 5,
                "tick": 2,
                "value": CloseFloat(0.75),
            },
            {
                "id": 1,
                "tick": 3,
                "value": CloseFloat(0.5),
            },
            {
                "id": 8,
                "tick": 4,
                "value": CloseFloat(0.15),
            },
        ])

    def test_moves_multiple_points_to_right_overwrite(self):
        automation_value = AutomationValue()
        self.assertEqual(points_to_objects(automation_value.get_points()), [])
        automation_value.add_point(tick=3, value=0.5)
        automation_value.add_point(tick=1, value=0.25)
        automation_value.add_point(tick=1, value=0.65)
        automation_value.add_point(tick=1, value=0.95)
        automation_value.add_point(tick=2, value=0.75)
        automation_value.add_point(tick=2, value=0.85)
        automation_value.add_point(tick=2, value=0.35)
        automation_value.add_point(tick=4, value=0.15)
        self.assertEqual(points_to_objects(automation_value.get_points()), [
            {
                "id": 4,
                "tick": 1,
                "value": CloseFloat(0.95),
            },
            {
                "id": 3,
                "tick": 1,
                "value": CloseFloat(0.65),
            },
            {
                "id": 2,
                "tick": 1,
                "value": CloseFloat(0.25),
            },
            {
                "id": 7,
                "tick": 2,
                "value": CloseFloat(0.35),
            },
            {
                "id": 6,
                "tick": 2,
                "value": CloseFloat(0.85),
            },
            {
                "id": 5,
                "tick": 2,
                "value": CloseFloat(0.75),
            },
            {
                "id": 1,
                "tick": 3,
                "value": CloseFloat(0.5),
            },
            {
                "id": 8,
                "tick": 4,
                "value": CloseFloat(0.15),
            },
        ])
        points = automation_value.get_points()
        automation_value.move_points([points[1].id, points[2].id, points[3].id, points[4].id], 2, -0.3)
        self.assertEqual(points_to_objects(automation_value.get_points()), [
            {
                "id": 4,
                "tick": 1,
                "value": CloseFloat(0.95),
            },
            {
                "id": 3,
                "tick": 3,
                "value": CloseFloat(0.35),
            },
            {
                "id": 2,
                "tick": 3,
                "value": 0,
            },
            {
                "id": 7,
                "tick": 4,
                "value": CloseFloat(0.05),
            },
            {
                "id": 6,
                "tick": 4,
                "value": CloseFloat(0.55),
            },
            {
                "id": 8,
                "tick": 4,
                "value": CloseFloat(0.15),
            },
        ])

    def test_moves_single_point_no_overwrite(self):
        automation_value = AutomationValue()
        self.assertEqual(points_to_objects(automation_value.get_points()), [])
        automation_value.add_point(tick=3, value=0.5)
        automation_value.add_point(tick=1, value=0.25)
        automation_value.add_point(tick=2, value=0.75)
        automation_value.add_point(tick=4, value=0.15)
        self.assertEqual(points_to_objects(automation_value.get_points()), [
            {
                "id": 2,
                "tick": 1,
                "value": CloseFloat(0.25),
            },
            {
                "id": 3,
                "tick": 2,
                "value": CloseFloat(0.75),
            },
            {
                "id": 1,
                "tick": 3,
                "value": CloseFloat(0.5),
            },
            {
                "id": 4,
                "tick": 4,
                "value": CloseFloat(0.15),
            },
        ])
        points = automation_value.get_points()
        automation_value.move_points([points[2].id], -2, 0.3,  overwrite_values_in_drag_area=False)
        self.assertEqual(points_to_objects(automation_value.get_points()), [
            {
                "id": 2,
                "tick": 1,
                "value": CloseFloat(0.25),
            },
            {
                "id": 1,
                "tick": 1,
                "value": CloseFloat(0.8),
            },
            {
                "id": 3,
                "tick": 2,
                "value": CloseFloat(0.75),
            },

            {
                "id": 4,
                "tick": 4,
                "value": CloseFloat(0.15),
            },
        ])

    def test_moves_single_point_overwrite(self):
        automation_value = AutomationValue()
        self.assertEqual(points_to_objects(automation_value.get_points()), [])
        automation_value.add_point(tick=3, value=0.5)
        automation_value.add_point(tick=1, value=0.25)
        automation_value.add_point(tick=1, value=0.65)
        automation_value.add_point(tick=1, value=0.85)
        automation_value.add_point(tick=2, value=0.75)
        automation_value.add_point(tick=4, value=0.15)
        self.assertEqual(points_to_objects(automation_value.get_points()), [
            {
                "id": 4,
                "tick": 1,
                "value": CloseFloat(0.85),
            },
            {
                "id": 3,
                "tick": 1,
                "value": CloseFloat(0.65),
            },
            {
                "id": 2,
                "tick": 1,
                "value": CloseFloat(0.25),
            },
            {
                "id": 5,
                "tick": 2,
                "value": CloseFloat(0.75),
            },
            {
                "id": 1,
                "tick": 3,
                "value": CloseFloat(0.5),
            },
            {
                "id": 6,
                "tick": 4,
                "value": CloseFloat(0.15),
            },
        ])
        points = automation_value.get_points()
        automation_value.move_points([points[1].id], 3, -0.3)
        self.assertEqual(points_to_objects(automation_value.get_points()), [
            {
                "id": 4,
                "tick": 1,
                "value": CloseFloat(0.85),
            },
            {
                "id": 3,
                "tick": 4,
                "value": CloseFloat(0.35),
            },
            {
                "id": 6,
                "tick": 4,
                "value": CloseFloat(0.15),
            },
        ])

    def test_moves_points_in_range_no_overwrite(self):
        automation_value = AutomationValue()
        self.assertEqual(points_to_objects(automation_value.get_points()), [])
        automation_value.add_point(tick=3, value=0.5)
        automation_value.add_point(tick=1, value=0.25)
        automation_value.add_point(tick=2, value=0.75)
        automation_value.add_point(tick=4, value=0.15)
        self.assertEqual(points_to_objects(automation_value.get_points()), [
            {
                "id": 2,
                "tick": 1,
                "value": CloseFloat(0.25),
            },
            {
                "id": 3,
                "tick": 2,
                "value": CloseFloat(0.75),
            },
            {
                "id": 1,
                "tick": 3,
                "value": CloseFloat(0.5),
            },
            {
                "id": 4,
                "tick": 4,
                "value": CloseFloat(0.15),
            },
        ])
        automation_value.move_points_in_range(2, 3, 4, 0, overwrite_values_in_drag_area=False)
        self.assertEqual(points_to_objects(automation_value.get_points()), [
            {
                "id": 2,
                "tick": 1,
                "value": CloseFloat(0.25),
            },
            {
                "id": 4,
                "tick": 4,
                "value": CloseFloat(0.15),
            },
            {
                "id": 3,
                "tick": 6,
                "value": CloseFloat(0.75),
            },
            {
                "id": 1,
                "tick": 7,
                "value": CloseFloat(0.5),
            },
        ])


class AutomationDataTestCase(unittest.TestCase):
    def test_adds_automation(self):
        song, track = create_song()
        target = AutomationTarget(
            AutomationTargetType.AUDIO_PLUGIN,
            'pluginId1',
            'paramId1',
        )
        self.assertEqual(list(track.get_automation().get_automation_targets()), [])
        self.assertIsNone(
            track.get_automation().get_automation_value_by_id(target.to_tf_automation_target_id()),
        )

        track.get_automation().add_automation(target)
        self.assertEqual(len(list(track.get_automation().get_automation_targets())), 1)
        self.assertIsNotNone(track.get_automation().get_automation_value_by_target(target))
        self.assertTrue(track.get_automation().get_automation_value_by_id(
            target.to_tf_automation_target_id()) == track.get_automation().get_automation_value_by_target(target))

    def test_removes_automation(self):
        song, track = create_song()
        target = AutomationTarget(
            AutomationTargetType.AUDIO_PLUGIN,
            'pluginId1',
            'paramId1',
        )

        track.get_automation().add_automation(target)
        track.get_automation().add_automation(target)
        self.assertEqual(len(list(track.get_automation().get_automation_targets())), 2)
        self.assertIsNotNone(track.get_automation().get_automation_value_by_target(target))
        track.get_automation().remove_automation(target)
        self.assertEqual(len(list(track.get_automation().get_automation_targets())), 0)
        self.assertIsNone(track.get_automation().get_automation_value_by_target(target))

    def test_multiple_automation_targets_update_the_same_value(self):
        song, track = create_song()
        target1 = AutomationTarget(
            AutomationTargetType.AUDIO_PLUGIN,
            'pluginId1',
            'paramId1',
        )
        target2 = AutomationTarget(
            AutomationTargetType.AUDIO_PLUGIN,
            'pluginId1',
            'paramId1',
        )

        track.get_automation().add_automation(target1)
        track.get_automation().add_automation(target2)
        self.assertEqual(len(list(track.get_automation().get_automation_targets())), 2)
        self.assertFalse(
            track.get_automation().get_automation_value_by_target(target1).get_disabled(),  # type:ignore
        )
        self.assertFalse(track.get_automation().get_automation_value_by_target(target2).get_disabled())  # type:ignore
        track.get_automation().get_automation_value_by_target(target1).set_disabled(True)  # type:ignore
        self.assertTrue(track.get_automation().get_automation_value_by_target(target2).get_disabled())  # type:ignore


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from tuneflow_py.models.protos import song_pb2


class TempoEvent:
    def __init__(
            self, ticks: int | None = None, bpm: float | None = None, time: float | None = None, proto: song_pb2.TempoEvent |
            None = None):
        if proto is not None:
            self._proto = proto
        else:
            self._proto = song_pb2.TempoEvent()
            self._proto.ticks = ticks
            self._proto.bpm = bpm
            if time is not None:
                self._proto.time = time

    def get_ticks(self) -> int:
        return self._proto.ticks

    def set_ticks(self, ticks: int):
        self._proto.ticks = ticks

    def get_bpm(self) -> float:
        return self._proto.bpm

    def get_time(self) -> float:
        return self._proto.time

    def __repr__(self) -> str:
        return str(self._proto)



from __future__ import annotations
import re
from typing import Callable, List, Optional, Generator
import unicodedata

from tuneflow_py.utils import lower_equal
from tuneflow_py.models.protos import song_pb2
from tuneflow_py.models.song import Song

LyricTokenizer = Callable[[str], List[str]]
DEFAULT_PPQ = Song.get_default_resolution()


class LyricWord:
    '''
    LyricWord is the primary unit of a LyricLine
    The class is a wrapper of the LyricWord proto object
    A word is located with start_tick and end_tick
    '''
    # The default word placeholer for an empty lyric line
    PLACEHOLDER_WORD = "^%%^"

    def __init__(
        self,
        line: LyricLine,
        word: str | None = None,
        start_tick: int | None = None,
        end_tick: int | None = None,
        proto: song_pb2.LyricLine.LyricWord | None = None
    ):
        self.line = line
        if proto is not None:
            self._proto = proto
        else:
            self._proto = song_pb2.LyricLine.LyricWord()
            self._proto.word = word
            self._proto.start_tick = start_tick
            self._proto.end_tick = end_tick

    def get_line(self):
        return self.line

    def get_word(self):
        return self._proto.word

    def set_word(self, word: str):
        if not word:
            raise ValueError("Word cannot be empty")
        self._proto.word = word

    def get_start_tick(self):
        return self._proto.start_tick

    def set_start_tick(self, start_tick: int, resolve_order=True):
        '''
        Set a new start tick for the word.
        This might delete the word if the new start tick is greater than or equal to the end tick.

        Args:
            start_tick (int): The new start tick of the word.
        '''
        if start_tick >= self._proto.end_tick and resolve_order:
            self._delete_from_parent()
            return
        self._proto.start_tick = start_tick
        if resolve_order:
            self.line.sort_words()

    def get_end_tick(self):
        return self._proto.end_tick

    def set_end_tick(self, end_tick: int, resolve_order=True):
        '''
        Set a new end tick for the word.
        This might delete the word if the new end tick is less than or equal to the start tick.

        Args:
            end_tick (int): The new end tick of the word.
        '''
        if end_tick <= self._proto.start_tick and resolve_order:
            self._delete_from_parent()
            return
        self._proto.end_tick = end_tick

    def move_to(self, start_tick: int, end_tick: int):
        if start_tick >= end_tick:
            self._delete_from_parent()
        self._proto.start_tick = start_tick
        self._proto.end_tick = end_tick
        self.line.sort_words()

    def _delete_from_parent(self):
        '''
        Delete the word from the parent line.
        '''
        self.line.remove_word(self)

    def __eq__(self, other: LyricWord):
        return (
            self._proto.start_tick == other._proto.start_tick
            and self._proto.end_tick == other._proto.end_tick
            and self._proto.word == other._proto.word
        )


class LyricLine:
    '''
    Lyrics is composed of a list of LyricLine objects
    '''

    def __init__(
        self,
        lyrics: Lyrics,
        start_tick: int | None = None,
        proto: song_pb2.LyricLine | None = None
    ):
        self.lyrics = lyrics
        if proto is not None:
            self._proto = proto
        else:
            self._proto = song_pb2.LyricLine()

        if len(self._proto.words) == 0:
            # Create a default placeholder for the empty line
            self._proto.words.add(
                word=LyricWord.PLACEHOLDER_WORD,
                start_tick=start_tick,
                end_tick=start_tick + DEFAULT_PPQ,
            )

    def __len__(self) -> int:
        if self.is_empty():
            return 0
        return len(self._proto.words)

    def __getitem__(self, index: int) -> LyricWord:
        return self.get_word_at_index(index)

    def is_empty(self) -> bool:
        # A line is empty if it only contains a placeholder word
        return len(self._proto.words) == 1 and self._proto.words[0].word == LyricWord.PLACEHOLDER_WORD

    def get_words(self) -> Generator:
        for word_proto in self._proto.words:
            yield LyricWord(line=self, proto=word_proto)

    def get_sentence(self) -> str:
        if self.is_empty():
            return ""
        return "".join(word_proto.word for word_proto in self._proto.words)

    @staticmethod
    def _get_start_tick(line_proto: song_pb2.LyricLine) -> int:
        if len(line_proto.words) == 0:
            raise ValueError("Words cannot be empty")
        return line_proto.words[0].start_tick

    @staticmethod
    def _get_end_tick(line_proto: song_pb2.LyricLine) -> int:
        if len(line_proto.words) == 0:
            raise ValueError("Words cannot be empty")
        return max(line_proto.words, key=lambda word_proto: word_proto.end_tick).end_tick

    def get_start_tick(self) -> int:
        return LyricLine._get_start_tick(self._proto)

    def get_end_tick(self) -> int:
        return LyricLine._get_end_tick(self._proto)

    def move_to(self, start_tick: int, end_tick: int) -> None:
        original_start = self.get_start_tick()
        original_end = self.get_end_tick()
        if original_start is None or original_end is None:
            return
        start_tick = max(start_tick, 0)
        end_tick = max(end_tick, start_tick)
        ratio = round((end_tick - start_tick) / (original_end - original_start))
        for word in self.get_words():
            word_start_tick = round(
                start_tick + (word.get_start_tick() - original_start) * ratio
            )
            word_end_tick = round(
                start_tick + (word.get_end_tick() - original_start) * ratio
            )
            word.move_to(word_start_tick, word_end_tick)

    def _set_words_from_tokens(
        self,
        tokens: List[str],
        start_tick: int,
        end_tick: int,
    ):
        if len(tokens) == 0:
            self.clear()
            return
        tick_per_word = (end_tick - start_tick) // len(tokens)
        del self._proto.words[:]
        for i, word in enumerate(tokens):
            start = start_tick + i * tick_per_word
            # Adjust the last word to match the end tick
            end = start_tick + (i + 1) * tick_per_word if i < len(tokens) - 1 else end_tick
            self._proto.words.add(word=word, start_tick=start, end_tick=end)
        self.sort_words()

    def set_words_from_string(
        self,
        input: str,
        start_tick: int | None = None,
        end_tick: int | None = None,
        tokenizer: Optional[LyricTokenizer] = None,
    ):
        '''
        Update lyric word objects of the line from a string of words
        The string is split into words with a tokenizer if provided
        Otherwise, the default tokenizer is used

        Args:
            input (str): The string of words
            start_tick (int | None): The start tick of the line. Use the original tick if set None
            end_tick (int | None): The end tick of the line. Use the original tick if set None
            tokenizer (Optional[LyricTokenizer]): The tokenizer to use
        '''
        start_tick = self.get_start_tick() if start_tick is None else start_tick
        end_tick = self.get_end_tick() if end_tick is None else end_tick
        tokens = (
            tokenizer(input) if tokenizer
            else LyricLine.default_lyric_tokenizer(input)
        )
        if len(tokens) == 0:
            self.clear()
        self._set_words_from_tokens(tokens, start_tick, end_tick)

    @staticmethod
    def _get_index_of_word_at_tick(line: song_pb2.LyricLine, tick: int):
        '''
        Internal generator implementation that locates word at a given tick.
        '''
        if len(line.words) == 0 or tick < LyricLine._get_start_tick(line) or tick >= LyricLine._get_end_tick(line):
            return -1
        index = lower_equal(
            line.words,
            tick,
            key=lambda item: item if isinstance(item, int) else item.start_tick
        )
        if index in range(len(line.words)):
            return index
        return -1

    def get_index_of_word_at_tick(self, tick: int):
        '''
        Return the index of the word in this proto that contains the given tick,
        or -1 if no such word exists.

        Args:
            tick (int): The tick to search for.

        Return:
            int: The index of the line that contains the tick, or -1 if not found.
        '''
        return LyricLine._get_index_of_word_at_tick(self._proto, tick)

    def create_word(self, word: str, start_tick: int, end_tick: int, resolve_order: bool = True):
        if self.is_empty():
            # Remove the placeholder
            del self._proto.words[:]
        proto = self._proto.words.add(word=word, start_tick=start_tick, end_tick=end_tick)
        if resolve_order:
            self.sort_words()
        return LyricWord(line=self, proto=proto)

    def get_word_at_index(self, index: int) -> LyricWord:
        if index < 0 or index >= len(self._proto.words):
            raise IndexError("Index out of range")
        return LyricWord(
            line=self,
            proto=self._proto.words[index]
        )

    def clear(self):
        new_word = self._create_placeholder_word()
        del self._proto.words[:]
        self._proto.words.append(new_word._proto)

    def _create_placeholder_word(self):
        return LyricWord(
            line=self,
            word=LyricWord.PLACEHOLDER_WORD,
            start_tick=self.get_start_tick(),
            end_tick=self.get_start_tick() + DEFAULT_PPQ * 4 * 2,
        )

    def remove_word(self, word: LyricWord):
        search_index = lower_equal(self._proto.words, word._proto, key=lambda word: word.start_tick)
        while search_index >= 0:
            lyric_word = self._proto.words[search_index]
            if lyric_word == word._proto:
                self.remove_word_at_index(search_index)
                break
            search_index -= 1

    def remove_word_at_index(self, index: int):
        if index < 0 or index >= len(self._proto.words):
            raise IndexError("Index out of range")
        if len(self._proto.words) == 1:
            # Words will become empty, insert a default placeholder.
            self.clear()
        else:
            del self._proto.words[index]
            self.sort_words()

    def sort_words(self):
        self._proto.words.sort(key=lambda word: word.start_tick)
        self.lyrics.sort_lines()

    @staticmethod
    def default_lyric_tokenizer(input: str) -> List[str]:
        '''
        Tokenizes the input string based on specific rules.

        Args:
            input (str): The input string to be tokenized.

        Returns:
            List[str]: A list of tokens generated based on the rules.
        '''
        tokens = []

        current_token = ''
        for char in input:
            if char == ' ':
                # Handle spaces
                if current_token:
                    tokens.append(current_token)
                    current_token = ''
                tokens.append(char)
            elif re.match(r'[\u3040-\u30ff\u3400-\u4dbf\u4e00-\u9fff\uf900-\ufaff\uff66-\uff9f]', char):
                # Handle Japanese, Chinese, Korean characters
                if current_token:
                    tokens.append(current_token)
                    current_token = ''
                tokens.append(char)
            elif unicodedata.category(char).startswith('P'):
                # Handle punctuation marks
                if current_token:
                    if re.match(r"^[']+$", char) and re.match(r'^[A-Za-z0-9]+$', current_token):
                        current_token += char
                    else:
                        tokens.append(current_token)
                        tokens.append(char)
                        current_token = ''
                else:
                    tokens.append(char)
            else:
                current_token += char

        if current_token:
            tokens.append(current_token)

        return tokens


class Lyrics:
    def __init__(self, song: Song) -> None:
        self.song = song
        if song._proto.lyrics is None:
            song._proto.lyrics = song_pb2.Lyrics()
        self._proto = song._proto.lyrics

    def __getitem__(self, index: int) -> LyricLine:
        '''
        Access the lyric line at the given index.
        Usage: lyrics[index]
        '''
        return self.get_line_at_index(index)

    def __len__(self) -> int:
        '''
        Get the number of lyric lines.
        Usage: len(lyrics)
        '''
        return len(self._proto.lines)

    def create_line(self, start_tick: int, resolve_order: bool = True):
        '''
        Creates a new empty LyricLine object at the given start_tick.

        Args:
            start_tick (int): The starting tick for the new lyric line.
            resolve_order (bool, optional): Whether to sort the lyric lines by start_tick. Defaults to True.

        Returns:
            LyricLine: The newly created LyricLine object.
        '''
        line_proto = self._proto.lines.add()
        line = LyricLine(lyrics=self, start_tick=start_tick, proto=line_proto)
        if resolve_order:
            self.sort_lines()
        return line

    def create_line_from_string(self, input: str, start_tick: int, end_tick: int, tokenizer: Optional[LyricTokenizer] = None):
        '''
        Create a Line object from a string of words.

        Args:
            input (str): The input string of words.
            start_tick (int): The starting tick of the line.
            end_tick (int): The ending tick of the line.
            tokenizer (Optional[LyricTokenizer], optional): The tokenizer to use for splitting the input string into words.
                If None, the default tokenizer will be used. Defaults to None.

        Returns:
            Line: The created Line object.

        Raises:
            ValueError: If the input string is empty.
        '''
        line_proto = self._proto.lines.add()
        line = LyricLine(lyrics=self, start_tick=start_tick, proto=line_proto)
        if not input:
            return line
        line.set_words_from_string(input, start_tick, end_tick, tokenizer)
        return line

    @staticmethod
    def _get_index_of_line_at_tick(proto: song_pb2.Lyrics, tick: int):
        '''
        Internal generator implementation that locates lyrlic line at a given tick.
        '''
        if len(proto.lines) == 0:
            return -1
        start_search_index = lower_equal(
            proto.lines,
            tick,
            key=lambda item: item if isinstance(item, int) else LyricLine._get_start_tick(item)
        )
        for index in range(start_search_index, -1, -1):
            line = proto.lines[index]
            if tick >= LyricLine._get_start_tick(line) and tick <= LyricLine._get_end_tick(line):
                yield index
        return -1

    def get_index_of_line_at_tick(self, tick: int):
        '''
        Yield the index of the line in this proto that contains the given tick,
        or -1 if no such line exists.

        Args:
            tick (int): The tick to search for.

        Yields:
            int: The index of the line that contains the tick, or -1 if not found.
        '''
        for index in Lyrics._get_index_of_line_at_tick(self._proto, tick):
            yield index

    def get_line_at_index(self, index: int):
        if index < 0 or index >= len(self._proto.lines):
            raise IndexError("Index out of range")
        return LyricLine(lyrics=self, proto=self._proto.lines[index])

    def get_lines(self):
        for line_proto in self._proto.lines:
            if len(line_proto.words) == 0:
                raise Exception("Lyric line has no words")
            yield LyricLine(
                lyrics=self,
                proto=line_proto
            )

    def remove_line(self, line: LyricLine):
        search_index = lower_equal(
            self._proto.lines,
            line._proto,
            key=lambda line: LyricLine._get_start_tick(line)
        )
        while search_index >= 0:
            line_proto = self._proto.lines[search_index]
            if line_proto == line._proto:
                self.remove_line_at_index(search_index)
            search_index -= 1

    def remove_line_at_index(self, index: int):
        if index < 0 or index >= len(self._proto.lines):
            return
        del self._proto.lines[index]

    def clone_line(self, original_line: LyricLine):
        new_proto = self._proto.lines.add()
        new_proto.CopyFrom(original_line._proto)
        self.sort_lines()
        return LyricLine(lyrics=self, proto=new_proto)

    def sort_lines(self):
        self._proto.lines.sort(key=lambda line: LyricLine._get_start_tick(line))

    def clear(self):
        del self._proto.lines[:]



from tuneflow_py.models.song import Song, TrackType
from tuneflow_py.models.note import Note
from tuneflow_py.models.clip import Clip
from tuneflow_py.models.audio_plugin import AudioPlugin
from tuneflow_py.utils import db_to_volume_value
from typing import List
import unittest


def create_song():
    song = Song()
    song.create_tempo_change(ticks=1440, bpm=60)
    return song


def assert_notes_are_equal(notes1: List[Note], notes2: List[Note]):
    if len(notes1) != len(notes2):
        raise Exception(
            f'Notes are not of equal length, {len(notes1)} vs {len(notes2)}')
    for i in range(0, len(notes1)):
        if not notes1[i].equals(notes2[i]) or notes1[i].get_id() != notes2[i].get_id():
            raise Exception(
                f'{i}th notes of the two lists are not equal, \n{str(notes1[i])}\n vs \n{str( notes2[i],)}',
            )


def create_test_notes(note_specs: List, clip: Clip):
    return [Note(**spec, clip=clip) for spec in note_specs]


class TestCreateTracks(unittest.TestCase):
    def test_create_midi_track(self):
        song = create_song()
        track = song.create_track(type=TrackType.MIDI_TRACK)
        self.assertAlmostEqual(track.get_volume(), db_to_volume_value(0))
        self.assertEqual(track.get_pan(), 0)
        self.assertEqual(track.get_solo(), False)
        self.assertEqual(track.get_muted(), False)
        self.assertEqual(track.get_type(), TrackType.MIDI_TRACK)

    def test_create_audio_track(self):
        song = create_song()
        track = song.create_track(type=TrackType.AUDIO_TRACK)
        self.assertAlmostEqual(track.get_volume(), db_to_volume_value(0))
        self.assertEqual(track.get_pan(), 0)
        self.assertEqual(track.get_solo(), False)
        self.assertEqual(track.get_muted(), False)
        self.assertEqual(track.get_type(), TrackType.AUDIO_TRACK)

    def test_create_track_returns_correct_reference(self):
        song = create_song()
        self.assertEqual(song.get_track_count(), 0)
        new_track = song.create_track(type=TrackType.MIDI_TRACK)
        self.assertEqual(new_track.get_instrument().program, 0)  # type:ignore
        self.assertEqual(song.get_track_at(0).get_instrument().program, 0)  # type:ignore
        new_track.get_instrument().program = 127  # type:ignore
        self.assertEqual(new_track.get_instrument().program, 127)  # type:ignore
        self.assertEqual(song.get_track_at(0).get_instrument().program, 127)  # type:ignore


class TestAudioTracks(unittest.TestCase):
    def test_cannot_set_instrument_or_suggested_instrument_or_sampler_plugin(self):
        song = create_song()
        track = song.create_track(type=TrackType.AUDIO_TRACK)
        self.assertFalse(track.has_instrument())
        self.assertFalse(track.has_sampler_plugin())
        self.assertEqual(track.get_suggested_instruments_count(), 0)
        track.set_instrument(program=32, is_drum=False)
        self.assertFalse(track.has_instrument())
        track.set_sampler_plugin(AudioPlugin(name="plugin1", manufacturer_name="manufacturer1",
                                 plugin_format_name="VST", plugin_version="1.1"))
        self.assertIsNone(track.get_sampler_plugin())
        self.assertFalse(track.has_sampler_plugin())
        track.create_suggested_instrument(program=64, is_drum=False)
        self.assertEqual(track.get_suggested_instruments_count(), 0)

    def test_clone_audio_track(self):
        # TODO: Test
        pass


class TestMIDITracks(unittest.TestCase):
    def test_clone_midi_track(self):
        # TODO: Test
        pass

    def test_get_visible_notes(self):
        song = create_song()
        track = song.create_track(type=TrackType.MIDI_TRACK)
        clip1 = track.create_midi_clip(clip_start_tick=10, clip_end_tick=20, insert_clip=True)
        clip2 = track.create_midi_clip(clip_start_tick=20, clip_end_tick=30, insert_clip=True)
        clip1.create_note(pitch=72, velocity=100, start_tick=10, end_tick=11)
        clip1.create_note(pitch=73, velocity=100, start_tick=11, end_tick=12)
        clip1.create_note(pitch=74, velocity=100, start_tick=12, end_tick=13)
        clip2.create_note(pitch=75, velocity=100, start_tick=20, end_tick=21)
        clip2.create_note(pitch=76, velocity=100, start_tick=21, end_tick=22)
        clip2.create_note(pitch=77, velocity=100, start_tick=22, end_tick=23)
        assert_notes_are_equal(track.get_visible_notes(), create_test_notes([
            {'pitch': 72, 'velocity': 100, 'start_tick': 10, 'end_tick': 11, 'id': 1},
            {'pitch': 73, 'velocity': 100, 'start_tick': 11, 'end_tick': 12, 'id': 2},
            {'pitch': 74, 'velocity': 100, 'start_tick': 12, 'end_tick': 13, 'id': 3},
            {'pitch': 75, 'velocity': 100, 'start_tick': 20, 'end_tick': 21, 'id': 1},
            {'pitch': 76, 'velocity': 100, 'start_tick': 21, 'end_tick': 22, 'id': 2},
            {'pitch': 77, 'velocity': 100, 'start_tick': 22, 'end_tick': 23, 'id': 3}
        ], clip1))


class TestCloneClip(unittest.TestCase):
    def test_clone_clip(self):
        song = create_song()
        track = song.create_track(type=TrackType.MIDI_TRACK)
        clip1 = track.create_midi_clip(clip_start_tick=10, clip_end_tick=20, insert_clip=True)
        clip1.create_note(pitch=72, velocity=100, start_tick=15, end_tick=16)
        clip2 = track.clone_clip(clip1)
        self.assertNotEqual(clip1.get_id(), clip2.get_id())
        self.assertEqual(clip2.get_clip_start_tick(), clip1.get_clip_start_tick())
        self.assertEqual(clip2.get_clip_end_tick(), clip1.get_clip_end_tick())
        self.assertEqual(clip2.get_raw_note_at(0).get_pitch(), 72)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from tuneflow_py.descriptors.clip_descriptor import AudioClipData
from tuneflow_py.models.protos import song_pb2
from tuneflow_py.models.note import Note
from tuneflow_py.utils import lower_than, greater_than, greater_equal
from nanoid import generate as generate_nanoid
from typing import List
from types import SimpleNamespace


ClipType = song_pb2.ClipType


class Clip:
    def __init__(self, song, type: int | None = None, clip_start_tick: int | None = None, id: str | None = None, track=None,
                 clip_end_tick: int | None = None, audio_clip_data: AudioClipData | None = None, proto: song_pb2.Clip |
                 None = None) -> None:
        '''
        A clip is a piece in a track, and it contains notes and the clip range.
        One track can contain one or many non-overlapping clips.

        IMPORTANT: Do not use the constructor directly, call `createClip` from tracks instead.

        @param song The song that the clip belongs to, should always be provided.
        @param track If the clip is dettached from any track, this field is going to be None.
        @param clip_start_tick Inclusive start tick.
        @param clip_end_tick Inclusive end tick.
        @param audio_clip_data Audio related data if the clip type is AUDIO_CLIP.
        '''
        self.song = song
        self.track = track
        self._next_note_id = None
        if proto is not None:
            self._proto = proto
            return
        self._proto = song_pb2.Clip()
        self._proto.id = id if id is not None else Clip._generate_clip_id()
        self._proto.type = type
        if type == ClipType.AUDIO_CLIP:
            if audio_clip_data is None:
                raise Exception(
                    'Audio clip data must be provided for audio clips.')

            if "audio_file_path" in audio_clip_data and audio_clip_data["audio_file_path"] is not None and audio_clip_data["audio_file_path"] != "":
                self._proto.audio_clip_data.audio_file_path = audio_clip_data["audio_file_path"]
            if "audio_data" in audio_clip_data and audio_clip_data["audio_data"] is not None:
                self._proto.audio_clip_data.audio_data.format = audio_clip_data["audio_data"]["format"]
                self._proto.audio_clip_data.audio_data.data = audio_clip_data["audio_data"]["data"]
            self._proto.audio_clip_data.start_tick = audio_clip_data["start_tick"]
            self._proto.audio_clip_data.duration = audio_clip_data["duration"]
            if "pitch_offset" in audio_clip_data:
                self._proto.audio_clip_data.pitch_offset = audio_clip_data["pitch_offset"]
            if "speed_ratio" in audio_clip_data:
                self._proto.audio_clip_data.speed_ratio = audio_clip_data["speed_ratio"]

            clip_start_tick = max(
                clip_start_tick, audio_clip_data["start_tick"]) if clip_start_tick is not None else audio_clip_data["start_tick"]
            audio_end_tick = self.get_audio_end_tick()
            if (clip_end_tick is None or audio_end_tick < clip_end_tick):  # type:ignore
                clip_end_tick = audio_end_tick

            self._proto.clip_start_tick = clip_start_tick
            self._proto.clip_end_tick = clip_end_tick
        elif (type == ClipType.MIDI_CLIP):
            self._proto.clip_start_tick = clip_start_tick
            if clip_end_tick is None:
                raise Exception(
                    'clip end tick must be provided when creating MIDI clip.')

            self._proto.clip_end_tick = clip_end_tick

    def get_id(self):
        return self._proto.id

    def get_track(self):
        return self.track

    def get_audio_end_tick(self) -> int | None:
        '''
        Gets the current clip audio's end tick.

        Returns None if the clip is not audio clip or audio clip data is missing.
        '''
        if self.get_type() != ClipType.AUDIO_CLIP or not self.has_audio_clip_data():
            return None

        duration = self.get_audio_duration()
        if duration is None:
            return None

        audio_start_time = self.song.tick_to_seconds(
            self._proto.audio_clip_data.start_tick)
        return self.song.seconds_to_tick(audio_start_time + duration)

    def get_duration(self):
        '''
        Duration from the clip start to clip end, in seconds.
        '''
        return self.song.tick_to_seconds(
            self.get_clip_end_tick()) - self.song.tick_to_seconds(
            self.get_clip_start_tick())

    def get_audio_duration(self) -> float | None:
        '''
        Gets the STRETCHED, FULL audio duration if the clip is AUDIO_CLIP, this takes speed ratio into consideration.

        Returns None if the clip is not AUDIO_CLIP or audio clip data is missing.
        '''
        if self.get_type() != ClipType.AUDIO_CLIP or not self.has_audio_clip_data():
            return None
        return self._proto.audio_clip_data.duration / self.get_audio_speed_ratio()

    def get_raw_audio_duration(self) -> float | None:
        '''
        Gets the UNSTRETCHED, FULL audio duration if the clip is AUDIO_CLIP, this DOES NOT consider speed ratio.

        Returns undefined if the clip is not AUDIO_CLIP or audio clip data is missing.
        '''
        if self.get_type() != ClipType.AUDIO_CLIP or not self.has_audio_clip_data():
            return None
        return self._proto.audio_clip_data.duration

    def get_audio_speed_ratio(self) -> float:
        if self.get_type() != ClipType.AUDIO_CLIP or not self.has_audio_clip_data() or self._proto.audio_clip_data.speed_ratio is None or self._proto.audio_clip_data.speed_ratio == 0:
            return 1
        return self._proto.audio_clip_data.speed_ratio

    def time_stretch_from_clip_left(self, to_left_tick: int, resolve_clip_conflict=True):
        '''
        Time-stretch the clip by adjusting the start tick of the clip.

        NOTE: This could delete the clip if the clip range becomes empty after
        this call.

        @param `to_left_tick` The new start tick to stretch the clip to.
        '''
        if (to_left_tick >= self.get_clip_end_tick()):
            self.delete_from_parent(delete_associated_track_automation=True)
            return
        stretch_factor = (self.get_clip_end_tick(
        ) - to_left_tick) / (self.get_clip_end_tick() - self.get_clip_start_tick())
        # Resolve conflict before changing the clip's range
        # to preserve the current order of clips.
        if (resolve_clip_conflict and self.track is not None):
            self.track._resolve_clip_conflict(
                self.get_id(),
                min(self.get_clip_start_tick(), to_left_tick),
                self.get_clip_end_tick(),
            )

        if self.get_type() == ClipType.MIDI_CLIP:
            self._time_stretch_midi_clip(stretch_factor=stretch_factor, reference_tick=self.get_clip_end_tick())
            self.move_clip_to(to_left_tick, move_associated_track_automation_points=False)
        elif (self.get_type() == ClipType.AUDIO_CLIP):
            self._time_stretch_audio_clip(to_left_tick, self.get_clip_end_tick(), self.get_clip_end_tick())
        # TODO: Scale the associated automation points accordingly.

    def time_stretch_from_clip_right(self, to_right_tick: int, resolve_clip_conflict=True):
        '''
        Time-stretch the clip by adjusting the end tick of the clip.

        NOTE: This could delete the clip if the range becomes empty after
        this call.
        @param to_right_tick The new end tick to stretch the clip to.
        '''
        if (to_right_tick <= self.get_clip_start_tick()):
            self.delete_from_parent(delete_associated_track_automation=True)
            return

        # Resolve conflict before changing the clip's range
        # to preserve the current order of clips.
        if resolve_clip_conflict and self.track:
            self.track._resolve_clip_conflict(
                self.get_id(),
                self.get_clip_start_tick(),
                max(self.get_clip_end_tick(), to_right_tick),
            )

        if (self.get_type() == ClipType.MIDI_CLIP):
            stretch_factor = (to_right_tick - self.get_clip_start_tick()
                              ) / (self.get_clip_end_tick() - self.get_clip_start_tick())
            self._time_stretch_midi_clip(stretch_factor=stretch_factor, reference_tick=self.get_clip_start_tick())
        elif (self.get_type() == ClipType.AUDIO_CLIP):
            self._time_stretch_audio_clip(self.get_clip_start_tick(), to_right_tick, self.get_clip_start_tick())

        # TODO: Scale the associated automation points accordingly.

    def _time_stretch_midi_clip(self, stretch_factor: float, reference_tick: int):
        for note in self.get_raw_notes():
            note.set_start_tick(
                Clip._calculate_scaled_new_tick(
                    note.get_start_tick(),
                    reference_tick, stretch_factor))
            note.set_end_tick(
                Clip._calculate_scaled_new_tick(
                    note.get_end_tick(),
                    reference_tick, stretch_factor))

        self._proto.clip_start_tick = Clip._calculate_scaled_new_tick(
            self.get_clip_start_tick(),
            reference_tick,
            stretch_factor,
        )
        self._proto.clip_end_tick = Clip._calculate_scaled_new_tick(
            self.get_clip_end_tick(), reference_tick, stretch_factor)

    def _time_stretch_audio_clip(
            self,
        new_left_tick: int,
        new_right_tick: int,
        reference_tick: int,
    ):
        if (self.get_type() != ClipType.AUDIO_CLIP):
            return

        audio_clip_data = self.get_audio_clip_data()
        if (audio_clip_data is None):
            raise Exception(f"Audio clip data is missing for audio clip {self.get_id()}")
        old_speed_ratio = audio_clip_data.speed_ratio if audio_clip_data.speed_ratio is not None and audio_clip_data.speed_ratio > 0 else 1
        old_duratoin = self.song.tick_to_seconds(
            self.get_clip_end_tick()) - self.song.tick_to_seconds(self.get_clip_start_tick())
        new_duration = self.song.tick_to_seconds(new_right_tick) - self.song.tick_to_seconds(new_left_tick)
        time_stretch_factor = new_duration / old_duratoin
        speed_ratio = old_speed_ratio / time_stretch_factor
        Clip.validate_audio_speed_ratio(speed_ratio)
        old_audio_start_time = self.song.tick_to_seconds(audio_clip_data.start_tick)
        reference_tick_time = self.song.tick_to_seconds(reference_tick)
        new_audio_start_to_reference_distance_in_time = (
            reference_tick_time - old_audio_start_time) * time_stretch_factor
        new_audio_start_time = reference_tick_time - new_audio_start_to_reference_distance_in_time
        new_audio_start_tick = self.song.seconds_to_tick(new_audio_start_time)
        audio_clip_data.speed_ratio = speed_ratio
        audio_clip_data.start_tick = new_audio_start_tick
        self._proto.clip_start_tick = new_left_tick
        self._proto.clip_end_tick = new_right_tick

    @staticmethod
    def validate_audio_speed_ratio(speed_ratio: float):
        if (
            speed_ratio is None or
            speed_ratio < Clip.MIN_AUDIO_SPEED_RATIO or
            speed_ratio > Clip.MAX_AUDIO_SPEED_RATIO
        ):
            raise Exception(
                f'Speed ratio must be >= {Clip.MIN_AUDIO_SPEED_RATIO} and <= {Clip.MAX_AUDIO_SPEED_RATIO}, you are changing to {speed_ratio}',
            )

    @staticmethod
    def _calculate_scaled_new_tick(
        old_tick: int,
        reference_tick: int,
        scale_factor: float,
    ):
        distance = (reference_tick - old_tick) * scale_factor
        return round(reference_tick - distance)

    def get_audio_pitch_offset(self):
        '''
        @returns The pitch offset of this audio in semitones, comparing to the raw audio.
        '''
        if (
            self.get_type() != ClipType.AUDIO_CLIP or
            not self.has_audio_clip_data() or
            self.get_audio_clip_data().pitch_offset is None  # type: ignore
        ):
            return 0
        return self.get_audio_clip_data().pitch_offset  # type: ignore

    def set_audio_pitch_offset(self, offset_in_semitones: float):
        '''
        Sets the pitch offset of this audio in semitones, the offset is compared to the raw audio.
        * @param offset_in_semitones Ranging from `Clip.MIN_AUDIO_PITCH_OFFSET` to `Clip.MAX_AUDIO_PITCH_OFFSET`, step 1, inclusive.
        * @returns
        '''
        if (self.get_type() != ClipType.AUDIO_CLIP):
            return

        Clip.validate_audio_pitch_offset(offset_in_semitones)
        self.get_audio_clip_data().pitch_offset = offset_in_semitones  # type: ignore

    @staticmethod
    def validate_audio_pitch_offset(offset_in_semitones: float):
        if (
            offset_in_semitones is None or
            offset_in_semitones < Clip.MIN_AUDIO_PITCH_OFFSET or
            offset_in_semitones > Clip.MAX_AUDIO_PITCH_OFFSET
        ):
            raise Exception(
                f'Pitch offset must be >= {Clip.MIN_AUDIO_PITCH_OFFSET} and <= {Clip.MAX_AUDIO_PITCH_OFFSET}, you are setting to {offset_in_semitones}',
            )

    def get_audio_start_tick(self):
        '''
        Gets the current clip audio's start tick.

        Note that this differs from the clip's start tick in that the user can trim
        the clip to only play part of the audio.

        Returns None if the clip is not audio clip or audio clip data is missing.
        '''
        if (self.get_type() != ClipType.AUDIO_CLIP or not self.has_audio_clip_data()):
            return None

        return self.get_audio_clip_data().start_tick  # type: ignore

    def get_raw_note_count(self):
        return len(self._proto.notes)

    def get_raw_notes(self):
        '''
        @returns All notes contained by the clip, including those that
        are not within the clip's range.
        '''
        for i in range(len(self._proto.notes)):
            yield self.get_raw_note_at(i)

    def get_raw_note_at(self, index: int):
        return Note(proto=self._proto.notes[index], clip=self)

    def get_notes(self):
        '''
        @returns Notes within the clip's range.
        '''

        note_protos = Clip._get_notes_in_range(
            raw_notes=self._proto.notes, start_tick=self.get_clip_start_tick(),
            end_tick=self.get_clip_end_tick())
        for note_proto in note_protos:
            yield Note(proto=note_proto, clip=self)

    def create_note(
        self,
        pitch: int,
        velocity: int,
        start_tick: int,
        end_tick: int,
        update_clip_range: bool = True,
        resolve_clip_conflict: bool = True
    ):
        '''
        Adds a note to the clip and returns it.

        @param pitch An integer value between 0 - 127
        @param velocity An integer value between 0 - 127
        @param start_tick An integer value indicating the start tick.
        @param end_tick An integer value indicating the end tick.
        @param update_clip_range Whether to update the clip's range if the note stretches outside the clip.
        @param resolve_clip_conflict Whether to resolve clip conflict if the clip range is updated.
        '''
        if self.get_type() != ClipType.MIDI_CLIP:
            # Only MIDI clips can create notes.
            return None
        if (
            not Note.is_valid_pitch(pitch) or
            not Note.is_note_range_valid(start_tick, end_tick) or
            not Note.is_note_velocity_valid(velocity)
        ):
            return None

        note = Note(
            pitch=pitch,
            velocity=velocity,
            start_tick=start_tick,
            end_tick=end_tick,
            id=self._get_next_note_id(),
        )
        if (start_tick < self.get_clip_start_tick() and update_clip_range):
            self.adjust_clip_left(start_tick, resolve_clip_conflict)

        if (end_tick > self.get_clip_end_tick() and update_clip_range):
            self.adjust_clip_right(end_tick, resolve_clip_conflict)

        self._ordered_insert_note(note)
        return note

    def get_type(self) -> int:
        return self._proto.type

    def delete_note(self, note: Note):
        note_index = self._get_note_index(note)
        if (note_index >= 0):
            self.delete_note_at(note_index)
            note.clip = None

    def delete_note_at(self, index: int):
        if (index >= 0 and index < len(self._proto.notes)):
            self._proto.notes.pop(index)

    def get_clip_start_tick(self) -> int:
        return self._proto.clip_start_tick

    def get_clip_end_tick(self) -> int:
        return self._proto.clip_end_tick

    def has_audio_clip_data(self) -> bool:
        return self._proto.HasField("audio_clip_data")

    def get_audio_clip_data(self) -> AudioClipData | None:
        if not self.has_audio_clip_data():
            return None
        return self._proto.audio_clip_data

    def set_audio_file(self, file_path: str, start_tick: int, duration: float):
        '''
        Sets a new audio file.
        '''
        self._proto.audio_clip_data.audio_file_path = file_path
        self._proto.audio_clip_data.start_tick = start_tick
        self._proto.audio_clip_data.duration = duration
        self._proto.audio_clip_data.speed_ratio = 1
        self._proto.audio_clip_data.pitch_offset = 0

    def clear_notes(self):
        del self._proto.notes[:]

    def delete_from_parent(self, delete_associated_track_automation: bool):
        if self.track is not None:
            self.track.delete_clip(self, delete_associated_track_automation)
            self.track = None

    def adjust_clip_left(self, clip_start_tick: int, resolve_conflict=True):
        '''
        Adjust the left boundary of the clip.

        NOTE: This could delete the clip if the range becomes empty after
        this call.

        @param clip_start_tick The new start tick (inclusive) of the clip.
        '''
        clip_start_tick = max(0, clip_start_tick)
        if self.get_type() == ClipType.AUDIO_CLIP and self.has_audio_clip_data():
            # Clip boundary cannot exceed the audio's boundary.
            clip_start_tick = max(
                clip_start_tick, self._proto.audio_clip_data.start_tick)

        if clip_start_tick > self.get_clip_end_tick():
            self.delete_from_parent(delete_associated_track_automation=True)
        else:
            # Resolve conflict before changing the clip's range
            # to preserve the current order of clips.
            if (resolve_conflict and self.track is not None):
                self.track._resolve_clip_conflict(
                    self.get_id(),
                    min(self.get_clip_start_tick(), clip_start_tick),
                    self.get_clip_end_tick(),
                )
            self._proto.clip_start_tick = clip_start_tick

    def adjust_clip_right(self, clip_end_tick: int, resolve_conflict=True):
        '''
        Adjust the right boundary of the clip.

        NOTE: This could delete the clip if the range becomes empty after
        this call.

        @param clipEndTick  The new end tick (inclusive) of the clip.
        '''
        if self.get_type() == ClipType.AUDIO_CLIP:
            # Clip right boundary cannot exceed the audio's right boundary.
            audio_end_tick = self.get_audio_end_tick()
            if audio_end_tick is not None:
                clip_end_tick = min(clip_end_tick, audio_end_tick)
        if clip_end_tick < self.get_clip_start_tick() or clip_end_tick < 0:
            self.delete_from_parent(delete_associated_track_automation=True)
        else:
            # Resolve conflict before changing the clip's range
            # to preserve the current order of clips.
            if (resolve_conflict and self.track is not None):
                self.track._resolve_clip_conflict(
                    self.get_id(),
                    self.get_clip_start_tick(),
                    max(self.get_clip_end_tick(), clip_end_tick),
                )

            self._proto.clip_end_tick = clip_end_tick

    def move_clip(self, offset_tick: int, move_associated_track_automation_points: bool):
        '''
        Move the clip by the given offset ticks.

        NOTE: Moving a clip can break the order of clips,
        so the clip must be re-positioned to the correct index.
        '''
        new_clip_start_tick = max(0, self.get_clip_start_tick() + offset_tick)
        new_clip_end_tick = self.get_clip_end_tick() + offset_tick
        if (new_clip_end_tick < 0):
            self.delete_from_parent(delete_associated_track_automation=True)
            return
        if self.track is not None:
            # Resolve conflict before changing the clip's range
            # to preserve the current order of clips.
            self.track._resolve_clip_conflict(self.get_id(), new_clip_start_tick, new_clip_end_tick)

        # The track's clip order will be invalidated after the move, and deleting the clip
        # requires the clips to be sorted, so delete the clip before the move happens.
        if self.track is not None:
            clip_index = self.track.get_clip_index(self)
            self.track.delete_clip_at(clip_index, delete_associated_track_automation=False)

        # Move the clip.
        original_start_tick = self.get_clip_start_tick()
        original_end_tick = self.get_clip_end_tick()

        if self.get_type() == ClipType.MIDI_CLIP:
            self._proto.clip_start_tick = new_clip_start_tick
            self._proto.clip_end_tick = new_clip_end_tick
            for note in self.get_raw_notes():
                note.set_start_tick(note.get_start_tick() + offset_tick)
                note.set_end_tick(note.get_end_tick() + offset_tick)

        elif (self.get_type() == ClipType.AUDIO_CLIP):
            if (not self.has_audio_clip_data()):
                raise Exception('Cannot move audio clip without audio data')

            # Let the content of the audio at the original clip start be S
            # Let the clip start tick before move be T1, after move be T2.
            # Moving an audio clip means:
            # 1. S must be at T2 after move
            # 2. Duration of the playable audio must remain unchanged.
            song = self.song
            # The clip's end position's time relative to the audio's start time should remain
            # unchanged.
            original_audio_start_time = song.tick_to_seconds(self.get_audio_clip_data().start_tick)  # type:ignore
            original_start_time = song.tick_to_seconds(original_start_tick)
            original_end_time = song.tick_to_seconds(original_end_tick)
            playable_audio_duration = original_end_time - original_start_time
            clip_start_relative_to_audio_start = original_start_time - original_audio_start_time
            # Use the raw updated clip tick here.
            new_clip_start_time = song.tick_to_seconds(self.get_clip_start_tick() + offset_tick)
            new_audio_start_time = new_clip_start_time - clip_start_relative_to_audio_start
            new_clip_end_time = new_clip_start_time + playable_audio_duration
            self._proto.clip_start_tick = new_clip_start_tick
            self._proto.clip_end_tick = song.seconds_to_tick(new_clip_end_time)
            self._proto.audio_clip_data.start_tick = song.seconds_to_tick(new_audio_start_time)
            if (self.get_clip_end_tick() < 0):
                self.delete_from_parent(delete_associated_track_automation=True)
                return

        if (self.track is not None):
            # Ordered insert back the clip.
            self.track._ordered_insert_clip(self)

            # Move track automation points if required.
            if (move_associated_track_automation_points):
                self.track.get_automation().move_all_points_within_range(
                    original_start_tick,
                    original_end_tick,
                    offset_tick,
                    offset_value=0,
                )

    def move_clip_to(self, tick: int, move_associated_track_automation_points: bool):
        '''
        Move the clip to a given tick.

        @param tick The tick that this clip will start at.
        '''
        offsetTick = tick - self.get_clip_start_tick()
        self.move_clip(offsetTick, move_associated_track_automation_points)

    @staticmethod
    def _get_notes_in_range(raw_notes: List[song_pb2.Note], start_tick: int, end_tick: int):
        '''
        Gets notes that are playable in a range.

        Performs two O(logn) searches and some extra while loops to narrow down range.
        '''
        # Find the first note that is within the clip.
        # IMPORTANT: Assuming that a note cannot be within the Clip if it starts before the clip.
        target_start_note = SimpleNamespace()
        target_start_note.start_tick = start_tick
        start_index = max(
            0,
            lower_than(
                raw_notes,
                target_start_note,
                key=lambda x: x.start_tick,
            ),
        )
        while (
            start_index >= 0 and start_index < len(raw_notes) and
            not Clip.is_note_in_clip(
                raw_notes[start_index].start_tick,
                raw_notes[start_index].end_tick,
                start_tick,
                end_tick,
            )
        ):
            start_index += 1

        if (start_index >= len(raw_notes)):
            return []

        # Find the last note that is within the clip.
        target_end_note = SimpleNamespace()
        target_end_note.start_tick = end_tick
        end_index = min(
            len(raw_notes) - 1,
            greater_than(
                raw_notes,
                target_end_note,
                lambda x: x.start_tick
            ),
        )
        while (
            end_index >= 0 and end_index < len(raw_notes) and
            not Clip.is_note_in_clip(
                raw_notes[end_index].start_tick,
                raw_notes[end_index].end_tick,
                start_tick,
                end_tick,
            )
        ):
            end_index -= 1

        if (end_index < 0):
            return []

        if (end_index < start_index):
            return []

        return raw_notes[start_index:(end_index + 1)]

    def _trim_conflict_part(self, start_tick: int, end_tick: int):
        '''
        Trim the conflict part from the clip.

        Anything within the given range (inclusive), will be removed from the current clip.
        '''
        overlapping_start_tick = max(start_tick, self.get_clip_start_tick())
        overlapping_end_tick = min(end_tick, self.get_clip_end_tick())
        if (overlapping_start_tick > overlapping_end_tick):
            # Overlapping range is empty.
            return
        if (
            overlapping_start_tick > self.get_clip_end_tick() or
            overlapping_end_tick < self.get_clip_start_tick()
        ):
            # No overlapping.
            return

        if (
            overlapping_end_tick >= self.get_clip_end_tick() and
            overlapping_start_tick <= self.get_clip_start_tick()
        ):
            # The whole clip overlaps with the given range.
            # Delete the clip.
            self.delete_from_parent(delete_associated_track_automation=True)
            return

        if (
            overlapping_end_tick < self.get_clip_end_tick() and
            overlapping_start_tick > self.get_clip_start_tick()
        ):
            # Overlapping part is in the middle of the clip.
            # Trim the clip into the left part and create a new clip for the right part.
            right_clip_start_tick = overlapping_end_tick + 1
            right_clip_end_tick = self.get_clip_end_tick()

            self.adjust_clip_right(
                overlapping_start_tick - 1, resolve_conflict=False)

            if self.track is not None:
                if self.get_type() == ClipType.MIDI_CLIP:
                    right_clip = self.track.create_midi_clip(
                        clip_start_tick=right_clip_start_tick,
                        clip_end_tick=right_clip_end_tick,
                    )

                    right_note_protos = Clip._get_notes_in_range(
                        raw_notes=self._proto.notes, start_tick=right_clip_start_tick, end_tick=right_clip_end_tick)
                    for right_note_proto in right_note_protos:
                        right_clip.create_note(
                            pitch=right_note_proto.pitch,
                            velocity=right_note_proto.velocity,
                            start_tick=right_note_proto.start_tick,
                            end_tick=right_note_proto.end_tick,
                            update_clip_range=False,
                            resolve_clip_conflict=False,
                        )

                elif (self.get_type() == ClipType.AUDIO_CLIP):
                    audio_clip_data = self._proto.audio_clip_data
                    self.track.create_audio_clip(
                        clip_start_tick=right_clip_start_tick,
                        clip_end_tick=right_clip_end_tick,
                        audio_clip_data={
                            "audio_file_path": audio_clip_data.audio_file_path if audio_clip_data.HasField("audio_file_path") else None,
                            "start_tick": audio_clip_data.start_tick,
                            "duration": audio_clip_data.duration,
                            "audio_data": {
                                "format": audio_clip_data.audio_data.format,
                                "data": audio_clip_data.audio_data.data
                            } if audio_clip_data.HasField("audio_data") else None
                        },
                    )
            return

        # Overlapping part is on the side.
        if (
            overlapping_start_tick > self.get_clip_start_tick() and
            overlapping_start_tick <= self.get_clip_end_tick()
        ):
            # Do not resolve conflict here since it will
            # cause a bad loop.
            self.adjust_clip_right(
                overlapping_start_tick - 1, resolve_conflict=False)
        elif (
            overlapping_end_tick < self.get_clip_end_tick() and
            overlapping_end_tick >= self.get_clip_start_tick()
        ):
            # Do not resolve conflict here since it will
            # cause a bad loop.
            self.adjust_clip_left(overlapping_end_tick +
                                  1, resolve_conflict=False)

    def _get_next_note_id(self):
        if self._next_note_id is None:
            if len(self._proto.notes) == 0:
                self._next_note_id = 1
            else:
                self._next_note_id = max([note_proto.id for note_proto in self._proto.notes]) + 1

        note_id = self._next_note_id
        if (self._next_note_id >= 2147483647):
            self._next_note_id = 1
        else:
            self._next_note_id += 1

        return note_id

    def _ordered_insert_note(self, new_note: Note):
        if (self.get_type() != ClipType.MIDI_CLIP):
            # Only MIDI clips can insert notes.
            return
        if (new_note.get_clip() == self):
            # Do not insert if the note is already in the list.
            return
        insert_index = greater_equal(
            self._proto.notes,
            new_note._proto,
            key=lambda x: x.start_tick,
        )
        if (insert_index < 0):
            self._proto.notes.append(new_note._proto)
            # Reassign proto since protobuf created a new copy
            new_note._proto = self._proto.notes[-1]
        else:
            self._proto.notes.insert(insert_index, new_note._proto)
            # Reassign proto since protobuf created a new copy
            new_note._proto = self._proto.notes[insert_index]
        new_note.clip = self

    def _get_note_index(self, note: Note):
        start_index = lower_than(
            self._proto.notes,
            note._proto,
            lambda x: x.start_tick,
        )

        for index in range(max(0, start_index), len(self._proto.notes)):
            current_note_proto = self._proto.notes[index]
            if current_note_proto.id == note.get_id():
                return index
            if (current_note_proto.start_tick > note.get_start_tick()):
                # We have searched past all possible notes.
                break
        return -1

    def __repr__(self) -> str:
        return str(self._proto)

    @staticmethod
    def is_note_in_clip(
        note_start_tick: int,
        note_end_tick: int,
        clip_start_tick: int,
        clip_end_tick: int,
    ):
        '''
        Note is in the clip when:
        * The note starts at or after the clip start.
        * The note has overlapping part with the clip.
        * The overlapping part is greater than 1 tick, i.e. the note does not start from the end of the clip or end at the start of the clip.
        '''
        # If the clip starts at 0, notes that start before 0 will start at 0.
        return (
            (note_start_tick >= clip_start_tick or (clip_start_tick == 0 and note_start_tick <= 0)) and
            note_start_tick < clip_end_tick and
            note_end_tick > note_start_tick
        )

    @staticmethod
    def _generate_clip_id():
        return generate_nanoid(size=10)

    MIN_AUDIO_SPEED_RATIO = 0.05
    MAX_AUDIO_SPEED_RATIO = 20
    MIN_AUDIO_PITCH_OFFSET = -24
    MAX_AUDIO_PITCH_OFFSET = 24
